"""
_grib_idx.py
============

This module provides utility functions for indexing and then utilizing
the index for the messages in GRIB2 files, accompanied by their idx files.

The module includes functions for extracting grib metadata, parsing the
attributes in idx files, one-to-one mapping between the grib metadata
and idx data, building the index for the grib files. It has been tested
with GFS, HRRR, GEFS model outputs.

Attributes
----------
(to be added)


Functions
---------
parse_grib_idx(basename, suffix, storage_options, validate)
    Parses the idx file with corresponding basename(grib file) with
    storage options, and returns the idx data as a Dataframe

extract_dataset_chunk_index(dset, ref_store, grib)
    Process and extracts the kerchunk index for an xarray dataset
    or datatree node and returns index meta data as a list of dictionary

extract_datatree_chunk_index(dtree, kerchunk_store, grib)
    Recursively iterate through the xarray datatree and with the
    help of extract_dataset_chunk_index, the index metadata as a Dataframe

build_idx_grib_mapping(basename, storage_options, suffix, mapper, validate)
    Creates a one-to-one mapping from the idx metadata and gribdata for a single
    file, which is needed to build the index. Mapping is returned as a Dataframe

map_from_index(run_time, mapping, idxdf, raw_merged)
    Builds the index for each grib file using the mapping made from
    building_idx_grib_mapping function and returns the index as a Dataframe.

Notes
-----
Indexing process should only happen for files belonging to the same horizon or else
the index won't work. This module is not be used directly but through kerchunk.grib2
module as the necessary functions for building the index are exposed there.


References
----------
For more information on indexing, check out this
link: https://github.com/asascience-open/nextgen-dmac/pull/57


Original license
----------------
MIT License Copyright (c) 2023 Camus Energy

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""  # noqa: E501

import ujson
import itertools
import os
import gzip
import re
import logging
import fsspec
from enum import unique, Enum
from functools import cache
import warnings
from typing import Iterable, Dict, Optional, Callable, Any
import numpy as np
import pandas as pd
import xarray as xr
import base64
import copy


class DynamicZarrStoreError(ValueError):
    pass


logger = logging.getLogger("grib-indexing")

COORD_DIM_MAPPING: dict[str, str] = dict(
    time="run_times",
    valid_time="valid_times",
    step="model_horizons",
)

ZARR_TREE_STORE = "zarr_tree_store.json.gz"

# derived from eccodes/grib2/typeOfLevelConcept.def
ECCODES_VERTICAL_LEVELS = {
    "abstractMultipleLevels",
    "abstractSingleLevel",
    "adiabaticCondensation",
    "atmosphere",
    "cloudBase",
    "cloudTop",
    "convectiveCondensation",
    "cumulonimbusBase",
    "cumulonimbusTop",
    "depthBelowLand",
    "depthBelowLandLayer",
    "depthBelowSea",
    "depthBelowSeaLayer",
    "entireAtmosphere",
    "entireLake",
    "entireMeltPond",
    "entireOcean",
    "eta",
    "freeConvection",
    "generalVertical",
    "generalVerticalLayer",
    "heightAboveGround",
    "heightAboveGroundLayer",
    "heightAboveSea",
    "heightAboveSeaLayer",
    "highCloudLayer",
    "hybrid",
    "hybridHeight",
    "hybridLayer",
    "hybridPressure",
    "iceBottomOnWater",
    "iceLayerAboveWaterSurface",
    "iceLayerOnWater",
    "iceLayerUnderSnowOnWater",
    "iceTopOnWater",
    "iceTopUnderSnowOnWater",
    "indefiniteSoilDepth",
    "isobaricInPa",
    "isobaricInhPa",
    "isobaricLayer",
    "isothermZero",
    "isothermal",
    "lakeBottom",
    "lowCloudLayer",
    "lowestLevelOfCloudCoverExceedance",
    "maxWind",
    "meanSea",
    "mediumCloudLayer",
    "meltPondBottom",
    "meltPondTop",
    "mixedLayerDepthByDensity",
    "mixedLayerDepthByDiffusivity",
    "mixedLayerDepthByTemperature",
    "mixedLayerDepthGeneric",
    "mixedLayerParcel",
    "mixingLayer",
    "mostUnstableParcel",
    "neutralBuoyancy",
    "nominalTop",
    "oceanModel",
    "oceanModelLayer",
    "oceanSurface",
    "oceanSurfaceToBottom",
    "potentialVorticity",
    "pressureFromGround",
    "pressureFromGroundLayer",
    "road",
    "roadLayer",
    "roof",
    "roofLayer",
    "rootZone",
    "seaBottom",
    "seaIce",
    "seaIceLayer",
    "sigma",
    "sigmaLayer",
    "snow",
    "snowLayer",
    "snowLayerOverIceOnWater",
    "snowTopOverIceOnWater",
    "soil",
    "soilLayer",
    "stratosphere",
    "surface",
    "theta",
    "thetaLayer",
    "totalSoilLayer",
    "tropopause",
    "troposphere",
    "unknown",
    "urbanCanyon",
    "wall",
    "wallLayer",
    "waterSurfaceToIsothermalOceanLayer",
}


def repeat_steps(step_index: pd.TimedeltaIndex, to_length: int) -> np.array:
    return np.tile(step_index.to_numpy(), int(np.ceil(to_length / len(step_index))))[
        :to_length
    ]


def create_steps(steps_index: pd.Index, to_length) -> np.array:
    return np.vstack([repeat_steps(si, to_length) for si in steps_index])


def store_coord_var(key: str, zstore: dict, coords: tuple[str, ...], data: np.array):
    if np.isnan(data).any():
        if f"{key}/.zarray" not in zstore:
            logger.debug("Skipping nan coordinate with no variable %s", key)
            return
        else:
            logger.info("Trying to add coordinate var %s with nan value!", key)

    zattrs = ujson.loads(zstore[f"{key}/.zattrs"])
    zarray = ujson.loads(zstore[f"{key}/.zarray"])
    # Use list not tuple
    zarray["chunks"] = [*data.shape]
    zarray["shape"] = [*data.shape]

    if key.endswith("step"):
        zarray["dtype"] = "<f8"
        zattrs["units"] = "hours"

    zattrs["_ARRAY_DIMENSIONS"] = [
        COORD_DIM_MAPPING[v] if v in COORD_DIM_MAPPING else v for v in coords
    ]

    zstore[f"{key}/.zarray"] = ujson.dumps(zarray)
    zstore[f"{key}/.zattrs"] = ujson.dumps(zattrs)

    vkey = ".".join(["0" for _ in coords])
    logger.info("COOORD %s DATA!: %s", key, data.dtype)
    data_bytes = data.tobytes()
    try:
        enocded_val = data_bytes.decode("ascii")
    except UnicodeDecodeError:
        enocded_val = (b"base64:" + base64.b64encode(data_bytes)).decode("ascii")
    zstore[f"{key}/{vkey}"] = enocded_val


def store_data_var(
    key: str,
    zstore: dict,
    dims: dict[str, int],
    coords: dict[str, tuple[str, ...]],
    data: pd.DataFrame,
    steps: np.array,
    times: np.array,
    lvals: Optional[np.array],
):
    zattrs = ujson.loads(zstore[f"{key}/.zattrs"])
    zarray = ujson.loads(zstore[f"{key}/.zarray"])

    dcoords = coords["datavar"]

    # The lat/lon y/x coordinates are always the last two
    lat_lon_dims = {
        k: v for k, v in zip(zattrs["_ARRAY_DIMENSIONS"][-2:], zarray["shape"][-2:])
    }
    full_coords = dcoords + tuple(lat_lon_dims.keys())
    full_dims = dict(**dims, **lat_lon_dims)

    # all chunk dimensions are 1 except for lat/lon or x/y
    zarray["chunks"] = [
        1 if c not in lat_lon_dims else lat_lon_dims[c] for c in full_coords
    ]
    zarray["shape"] = [full_dims[k] for k in full_coords]
    if zarray["fill_value"] is None:
        # Check dtype first?
        zarray["fill_value"] = np.nan

    zattrs["_ARRAY_DIMENSIONS"] = [
        COORD_DIM_MAPPING[v] if v in COORD_DIM_MAPPING else v for v in full_coords
    ]

    zstore[f"{key}/.zarray"] = ujson.dumps(zarray)
    zstore[f"{key}/.zattrs"] = ujson.dumps(zattrs)

    idata = data.set_index(["time", "step", "level"]).sort_index()

    for idx in itertools.product(*[range(dims[k]) for k in dcoords]):
        # Build an iterator over each of the single dimension chunks
        # TODO if the .loc call is slow inside the loop replace with a reindex operation and iterate that.
        dim_idx = {k: v for k, v in zip(dcoords, idx)}

        iloc: tuple[Any, ...] = (
            times[tuple([dim_idx[k] for k in coords["time"]])],
            steps[tuple([dim_idx[k] for k in coords["step"]])],
        )
        if lvals is not None:
            iloc = iloc + (lvals[idx[-1]],)  # type:ignore[assignment]

        try:
            # Squeeze if needed to get a series. Noop if already a series Df has multiple rows
            dval = idata.loc[iloc].squeeze()
        except KeyError:
            logger.info(f"Error getting vals {iloc} for in path {key}")
            continue

        assert isinstance(
            dval, pd.Series
        ), f"Got multiple values for iloc {iloc} in key {key}: {dval}"

        if pd.isna(dval.inline_value):
            # List of [URI(Str), offset(Int), length(Int)] using python (not numpy) types.
            record = [dval.uri, dval.offset.item(), dval.length.item()]
        else:
            record = dval.inline_value
        # lat/lon y/x have only the zero chunk
        vkey = ".".join([str(v) for v in (idx + (0, 0))])
        zstore[f"{key}/{vkey}"] = record


@unique
class AggregationType(Enum):
    """
    ENUM for aggregation types
    TODO is this useful elsewhere?
    """

    HORIZON = "horizon"
    VALID_TIME = "valid_time"
    RUN_TIME = "run_time"
    BEST_AVAILABLE = "best_available"


def reinflate_grib_store(
    axes: list[pd.Index],
    aggregation_type: AggregationType,
    chunk_index: pd.DataFrame,
    zarr_ref_store: dict,
) -> dict:
    """
    Given a zarr_store hierarchy, pull out the variables present in the chunks dataframe and reinflate the
    zarr variables adding any needed dimensions. This is a select operation - based on the time axis
    provided.
    Assumes everything is stored in hours per grib convention.

    :param axes: a list of new axes for aggregation
    :param aggregation_type: the type of fmrc aggregation
    :param chunk_index: a dataframe containing the kerchunk index
    :param zarr_ref_store: the deflated (chunks removed) zarr store
    :return: the inflated zarr store
    """
    # Make a deep copy to avoid modifying the input
    zstore = copy.deepcopy(zarr_ref_store["refs"])

    axes_by_name: dict[str, pd.Index] = {pdi.name: pdi for pdi in axes}
    # Validate axis names
    time_dims: dict[str, int] = {}
    time_coords: dict[str, tuple[str, ...]] = {}
    # TODO: add a data class or other method of typing and validating the variables created in this if block
    if aggregation_type.value == AggregationType.HORIZON.value:
        # Use index length horizons containing timedelta ranges for the set of steps
        time_dims["step"] = len(axes_by_name["step"])
        time_dims["valid_time"] = len(axes_by_name["valid_time"])

        time_coords["step"] = ("step", "valid_time")
        time_coords["valid_time"] = ("step", "valid_time")
        time_coords["time"] = ("step", "valid_time")
        time_coords["datavar"] = ("step", "valid_time")

        steps = create_steps(axes_by_name["step"], time_dims["valid_time"])
        valid_times = np.tile(
            axes_by_name["valid_time"].to_numpy(), (time_dims["step"], 1)
        )
        times = valid_times - steps

    elif aggregation_type.value == AggregationType.VALID_TIME.value:
        # Provide an index of steps and an index of valid times
        time_dims["step"] = len(axes_by_name["step"])
        time_dims["valid_time"] = len(axes_by_name["valid_time"])

        time_coords["step"] = ("step",)
        time_coords["valid_time"] = ("valid_time",)
        time_coords["time"] = ("valid_time", "step")
        time_coords["datavar"] = ("valid_time", "step")

        steps = axes_by_name["step"].to_numpy()
        valid_times = axes_by_name["valid_time"].to_numpy()

        steps2d = np.tile(axes_by_name["step"], (time_dims["valid_time"], 1))
        valid_times2d = np.tile(
            np.reshape(axes_by_name["valid_time"], (-1, 1)), (1, time_dims["step"])
        )
        times = valid_times2d - steps2d

    elif aggregation_type.value == AggregationType.RUN_TIME.value:
        # Provide an index of steps and an index of run times.
        time_dims["step"] = len(axes_by_name["step"])
        time_dims["time"] = len(axes_by_name["time"])

        time_coords["step"] = ("step",)
        time_coords["valid_time"] = ("time", "step")
        time_coords["time"] = ("time",)
        time_coords["datavar"] = ("time", "step")

        steps = axes_by_name["step"].to_numpy()
        times = axes_by_name["time"].to_numpy()

        # The valid times will be runtimes by steps
        steps2d = np.tile(axes_by_name["step"], (time_dims["time"], 1))
        times2d = np.tile(
            np.reshape(axes_by_name["time"], (-1, 1)), (1, time_dims["step"])
        )
        valid_times = times2d + steps2d

    elif aggregation_type.value == AggregationType.BEST_AVAILABLE.value:
        time_dims["valid_time"] = len(axes_by_name["valid_time"])
        assert (
            len(axes_by_name["time"]) == 1
        ), "The time axes must describe a single 'as of' date for best available"
        reference_time = axes_by_name["time"].to_numpy()[0]

        time_coords["step"] = ("valid_time",)
        time_coords["valid_time"] = ("valid_time",)
        time_coords["time"] = ("valid_time",)
        time_coords["datavar"] = ("valid_time",)

        valid_times = axes_by_name["valid_time"].to_numpy()
        times = np.where(valid_times <= reference_time, valid_times, reference_time)
        steps = valid_times - times
    else:
        raise RuntimeError(f"Invalid aggregation_type argument: {aggregation_type}")

    # Copy all the groups that contain variables in the chunk dataset
    unique_groups = chunk_index.set_index(
        ["varname", "stepType", "typeOfLevel"]
    ).index.unique()

    # Drop keys not in the unique groups
    for key in list(zstore.keys()):
        # Separate the key as a path keeping only: varname, stepType and typeOfLevel
        # Treat root keys like ".zgroup" as special and return an empty tuple
        lookup = tuple(
            [val for val in os.path.dirname(key).split("/")[:3] if val != ""]
        )
        if lookup not in unique_groups:
            del zstore[key]

    # Now update the zstore for each variable.
    for key, group in chunk_index.groupby(["varname", "stepType", "typeOfLevel"]):
        base_path = "/".join(key)
        lvals = group.level.unique()
        dims = time_dims.copy()
        coords = time_coords.copy()
        if len(lvals) == 1:
            lvals = lvals.squeeze()
            dims[key[2]] = 0
        elif len(lvals) > 1:
            lvals = np.sort(lvals)
            # multipel levels
            dims[key[2]] = len(lvals)
            coords["datavar"] += (key[2],)
        else:
            raise ValueError("")

        # Convert to floating point seconds
        # td.astype("timedelta64[s]").astype(float) / 3600  # Convert to floating point hours
        store_coord_var(
            key=f"{base_path}/time",
            zstore=zstore,
            coords=time_coords["time"],
            data=times.astype("datetime64[s]"),
        )

        store_coord_var(
            key=f"{base_path}/valid_time",
            zstore=zstore,
            coords=time_coords["valid_time"],
            data=valid_times.astype("datetime64[s]"),
        )

        store_coord_var(
            key=f"{base_path}/step",
            zstore=zstore,
            coords=time_coords["step"],
            data=steps.astype("timedelta64[s]").astype("float64") / 3600.0,
        )

        store_coord_var(
            key=f"{base_path}/{key[2]}",
            zstore=zstore,
            coords=(key[2],) if lvals.shape else (),
            data=lvals,  # all grib levels are floats
        )

        store_data_var(
            key=f"{base_path}/{key[0]}",
            zstore=zstore,
            dims=dims,
            coords=coords,
            data=group,
            steps=steps,
            times=times,
            lvals=lvals if lvals.shape else None,
        )

    return dict(refs=zstore, version=1)


def strip_datavar_chunks(
    kerchunk_store: dict, keep_list: tuple[str, ...] = ("latitude", "longitude")
) -> None:
    """
    Modify in place a kerchunk reference store to strip the kerchunk references for variables not in the
    keep list.
    :param kerchunk_store: a kerchunk ref spec store
    :param keep_list: the list of variables to keep references
    """
    zarr_store = kerchunk_store["refs"]

    zchunk_matcher = re.compile(r"^(?P<name>.*)\/(?P<zchunk>\d+[\.\d+]*)$")
    for key in list(zarr_store.keys()):
        matched = zchunk_matcher.match(key)
        if matched:
            logger.debug("Matched! %s", matched)
            if any([matched.group("name").endswith(keeper) for keeper in keep_list]):
                logger.debug("Skipping key %s", matched.group("name"))
                continue
            del zarr_store[key]


def write_store(metadata_path: str, store: dict):
    fpath = os.path.join(metadata_path, ZARR_TREE_STORE)
    compressed = gzip.compress(ujson.dumps(store).encode())
    with fsspec.open(fpath, "wb") as f:
        f.write(compressed)
    logger.info("Wrote %d bytes to %s", len(compressed), fpath)


@cache
def read_store(metadata_path: str) -> dict:
    """
    Cached method for loading the static zarr store from a metadata path
    :param metadata_path: the path (usually gcs) to the metadata directory
    :return: a kerchunk zarr store reference spec dictionary (defalated)
    """
    fpath = os.path.join(metadata_path, ZARR_TREE_STORE)
    with fsspec.open(fpath, "rb") as f:
        compressed = f.read()
    logger.info("Read %d bytes from %s", len(compressed), fpath)
    zarr_store = ujson.loads(gzip.decompress(compressed).decode())
    return zarr_store


def grib_coord(name: str) -> str:
    """
    Take advantage of gribs strict coordinate name structure
    Use the known level names and assume everything else is "level" which is described by the
    gribLevelAttribute
    This is helpful because there are a lot of named levels but they are all float values
    and each variable can have only one level coordinate.
    By mapping all of these levels to level the sparse chunk index becomes dense.
    :param name:
    :return:
    """

    if name in ("valid_time", "time", "step", "latitude", "longitude"):
        return name
    else:
        return "level"


def parse_grib_idx(
    basename: str,
    suffix: str = "idx",
    storage_options: Optional[Dict] = None,
    validate: bool = False,
) -> "pd.DataFrame":
    """
    Parses per-message metadata from a grib2.idx file (text-type) to a dataframe of attributes

    The function uses the idx file, extracts the metadata known as attrs (variables with
    level and forecast time) from each idx entry and converts it into pandas
    DataFrame. The dataframe is later to build the one-to-one mapping to the grib file metadata.

    Parameters
    ----------
    basename : str
        The base name is the full path to the grib file.
    suffix : str
        The suffix is the ending for the idx file.
    storage_options: dict
        For accessing the data, passed to filesystem
    validate : bool
        The validation if the metadata table has duplicate attrs.

    Returns
    -------
    pandas.DataFrame : The data frame containing the results.
    """
    import pandas as pd

    fs, _ = fsspec.core.url_to_fs(basename, **(storage_options or {}))

    fname = f"{basename}.{suffix}"

    baseinfo = fs.info(basename)

    with fs.open(fname) as f:
        result = pd.read_csv(f, header=None, names=["raw_data"])
        result[["idx", "offset", "date", "attrs"]] = result["raw_data"].str.split(
            ":", expand=True, n=3
        )
        result["offset"] = result["offset"].astype(int)
        result["idx"] = result["idx"].astype(int)

        # dropping the original single "raw_data" column after formatting
        result.drop(columns=["raw_data"], inplace=True)

    result = result.assign(
        length=(
            result.offset.shift(periods=-1, fill_value=baseinfo["size"]) - result.offset
        ),
        idx_uri=fname,
        grib_uri=basename,
    )

    if validate and not result["attrs"].is_unique:
        raise ValueError(f"Attribute mapping for grib file {basename} is not unique")

    return result.set_index("idx")


def build_path(path: Iterable[str | None], suffix: Optional[str] = None) -> str:
    """
    Returns the path to access the values in a zarr store without a leading "/"

    Parameters
    ----------
    path : Iterable[str | None]
        The path is the list of values to the element in zarr store
    suffix : str
        Last element if any

    Returns
    -------
    str : returns the path as a string
    """
    return "/".join([val for val in [*path, suffix] if val is not None]).lstrip("/")


def extract_dataset_chunk_index(
    dset: "xr.DataTree",
    ref_store: Dict,
    grib: bool = False,
) -> list[dict]:
    """
    Process and extracts the kerchunk index for an xarray dataset or datatree node.

    The data_vars from the dataset will be indexed.
    The coordinate vars for each dataset will be used for indexing.
    Datatrees generated by grib_tree have some nice properties which allow a denser index.

    Parameters
    ----------
    dset : datatree.DataTree
        The datatree node from the datatree instance
    ref_store : Dict
        The zarr store dictionary backed by the gribtree
    grib : bool
        boolean for treating coordinates as grib levels

    Returns
    -------
    list[dict] : returns the extracted grib metadata in the form of key-value pairs inside a list
    """

    result: list[dict] = []
    attributes = dset.attrs.copy()

    dpath = None
    if isinstance(dset, xr.DataTree):
        dpath = dset.path
        walk_group = dset.parent
        while walk_group:
            attributes.update(walk_group.attrs)
            walk_group = walk_group.parent

    for dname, dvar in dset.data_vars.items():
        # Get the chunk size - `chunks` property only works for xarray native
        zarray = ujson.loads(ref_store[build_path([dpath, dname], suffix=".zarray")])
        dchunk = zarray["chunks"]
        dshape = dvar.shape

        index_dims = {}
        for ddim_nane, ddim_size, dchunk_size in zip(dvar.dims, dshape, dchunk):
            if dchunk_size == 1:
                index_dims[ddim_nane] = ddim_size
            elif dchunk_size != ddim_size:
                # Must be able to get a single coordinate value for each chunk to index it.
                raise ValueError(
                    "Can not extract chunk index for dimension %s with non singleton chunk dimensions"
                    % ddim_nane
                )
            # Drop the dim where each chunk covers the whole dimension - no indexing needed!

        for idx in itertools.product(*[range(v) for v in index_dims.values()]):
            # Build an iterator over each of the single dimension chunks
            dim_idx = {key: val for key, val in zip(index_dims.keys(), idx)}

            coord_vals = {}
            for cname, cvar in dvar.coords.items():
                if grib:
                    # Grib data has only one level coordinate
                    cname = "level" if cname in ECCODES_VERTICAL_LEVELS else cname

                if all([dim_name in dim_idx for dim_name in cvar.dims]):
                    coord_index = tuple([dim_idx[dim_name] for dim_name in cvar.dims])
                    try:
                        coord_vals[cname] = cvar.to_numpy()[coord_index]
                    except Exception:
                        raise DynamicZarrStoreError(
                            f"Error reading coords for {dpath}/{dname} coord {cname} with index {coord_index}"
                        )

            whole_dim_cnt = len(dvar.dims) - len(dim_idx)
            chunk_idx = map(str, [*idx, *[0] * whole_dim_cnt])
            chunk_key = build_path([dpath, dname], suffix=".".join(chunk_idx))
            chunk_ref = ref_store.get(chunk_key)

            if chunk_ref is None:
                logger.warning("Chunk not found: %s", chunk_key)
                continue

            elif isinstance(chunk_ref, list) and len(chunk_ref) == 3:
                chunk_data = dict(
                    uri=chunk_ref[0],
                    offset=chunk_ref[1],
                    length=chunk_ref[2],
                    inline_value=None,
                )
            elif isinstance(chunk_ref, (bytes, str)):
                chunk_data = dict(inline_value=chunk_ref, offset=-1, length=-1)
            else:
                raise ValueError(f"Key {chunk_key} has bad value '{chunk_ref}'")
            result.append(dict(varname=dname, **attributes, **coord_vals, **chunk_data))

    return result


def extract_datatree_chunk_index(
    dtree: "xr.DataTree", kerchunk_store: dict, grib: bool = False
) -> "pd.DataFrame":
    """
    Recursive method to iterate over the data tree and extract the data variable chunks with index metadata

    Parameters
    ----------
    dtree : datatree.DataTree
        The xarray datatree representation of the reference filesystem
    kerchunk_store : dict
        the grib_tree output for a single grib file
    grib : bool
        boolean for treating coordinates as grib levels

    Returns
    -------
    pandas.Dataframe : The dataframe constructed from the grib metadata
    """
    import pandas as pd

    result: list[dict] = []

    for node in dtree.subtree:
        if node.has_data:
            result += extract_dataset_chunk_index(
                node, kerchunk_store["refs"], grib=grib
            )

    return pd.DataFrame.from_records(result)


def _map_grib_file_by_group(
    fname: str,
    mapper: Optional[Callable] = None,
    storage_options: Optional[Dict] = None,
) -> "pd.DataFrame":
    """
    Helper method used to read the cfgrib metadata associated with each message (group) in the grib file
    This method does not add metadata

    Parameters
    ----------
    fname : str
        the file name to read with scan_grib
    mapper : Optional[Callable]
        the mapper if any to apply (used for hrrr subhf)

    Returns
    -------
    pandas.Dataframe : The intermediate dataframe constructed from the grib metadata
    """
    import pandas as pd
    from kerchunk.grib2 import scan_grib

    mapper = (lambda x: x) if mapper is None else mapper
    references = scan_grib(fname, storage_options=storage_options)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return pd.concat(
            # grib idx is fortran indexed (from one not zero)
            list(
                filter(
                    # filtering out the cfgrib metadata dataframe in case it is None
                    lambda item: item is not None,
                    [
                        # extracting the metadata from a single message
                        _extract_single_group(mapper(group), i, storage_options)
                        for i, group in enumerate(references, start=1)
                    ],
                )
            )
        ).set_index("idx")


def _extract_single_group(grib_group: dict, idx: int, storage_options: Dict):
    from kerchunk.grib2 import grib_tree

    grib_tree_store = grib_tree(
        [
            grib_group,
        ],
        storage_options,
    )

    if len(grib_tree_store["refs"]) <= 1:
        logger.info("Empty DT: %s", grib_tree_store)
        return None

    dt = xr.open_datatree(
        "reference://",
        storage_options={"fo": grib_tree_store},
        engine="zarr",
        consolidated=False,
    )

    k_ind = extract_datatree_chunk_index(dt, grib_tree_store, grib=True)
    if k_ind.empty:
        logger.warning("Empty Kind: %s", grib_tree_store)
        return None

    assert (
        len(k_ind) == 1
    ), f"expected a single variable grib group but produced: {k_ind}"
    k_ind.loc[:, "idx"] = idx
    return k_ind


def build_idx_grib_mapping(
    basename: str,
    storage_options: Optional[Dict] = None,
    suffix: str = "idx",
    mapper: Optional[Callable] = None,
    validate: bool = True,
) -> "pd.DataFrame":
    """
    Mapping method combines the idx and grib metadata to make a mapping from
    one to the other for a particular model horizon file. This should be generally
    applicable to all forecasts for the given horizon.

    Parameters
    ----------
    basename : str
        the full path for the grib2 file
    storage_options: dict
        For accessing the data, passed to filesystem
    suffix : str
        The suffix is the ending for the idx file.
    mapper : Optional[Callable]
        the mapper if any to apply (used for hrrr subhf)
    validate : bool
        to assert the mapping is correct or fail before returning

    Returns
    -------
    pandas.Dataframe : The merged dataframe with the results of the two operations
    joined on the grib message (group) number
    """
    import pandas as pd

    grib_file_index = _map_grib_file_by_group(
        fname=basename,
        mapper=mapper,
        storage_options=storage_options,
    )
    idx_file_index = parse_grib_idx(
        basename=basename, suffix=suffix, storage_options=storage_options
    )
    result = idx_file_index.merge(
        # Left merge because the idx file should be authoritative - one record per grib message
        grib_file_index,
        on="idx",
        how="left",
        suffixes=("_idx", "_grib"),
    )

    if validate:
        # If any of these conditions fail - inspect the result manually on colab.
        all_match_offset = (
            (result.loc[:, "offset_idx"] == result.loc[:, "offset_grib"])
            | pd.isna(result.loc[:, "offset_grib"])
            | ~pd.isna(result.loc[:, "inline_value"])
        )
        all_match_length = (
            (result.loc[:, "length_idx"] == result.loc[:, "length_grib"])
            | pd.isna(result.loc[:, "length_grib"])
            | ~pd.isna(result.loc[:, "inline_value"])
        )

        if not all_match_offset.all():
            vcs = all_match_offset.value_counts()
            raise ValueError(
                f"Failed to match message offset mapping for grib file {basename}: "
                f"{vcs[True]} matched, {vcs[False]} didn't"
            )

        if not all_match_length.all():
            vcs = all_match_length.value_counts()
            raise ValueError(
                f"Failed to match message offset mapping for grib file {basename}: "
                f"{vcs[True]} matched, {vcs[False]} didn't"
            )

        if not result["attrs"].is_unique:
            dups = result.loc[result["attrs"].duplicated(keep=False), :]
            logger.warning(
                "The idx attribute mapping for %s is not unique for %d variables: %s",
                basename,
                len(dups),
                dups.varname.tolist(),
            )

        r_index = result.set_index(
            ["varname", "typeOfLevel", "stepType", "level", "valid_time"]
        )
        if not r_index.index.is_unique:
            dups = r_index.loc[r_index.index.duplicated(keep=False), :]
            logger.warning(
                "The grib hierarchy in %s is not unique for %d variables: %s",
                basename,
                len(dups),
                dups.index.get_level_values("varname").tolist(),
            )

    return result


def map_from_index(
    run_time: "pd.Timestamp",
    mapping: "pd.DataFrame",
    idxdf: "pd.DataFrame",
    raw_merged: bool = False,
) -> "pd.DataFrame":
    """
    Main method used for building index dataframes from parsed IDX files
    merged with the correct mapping for the horizon

    Parameters
    ----------

    run_time : pd.Timestamp
        the run time timestamp of the idx data
    mapping : pd.DataFrame
        the mapping data derived from comparing the idx attributes to the
        CFGrib attributes for a given horizon
    idxdf : pd.DataFrame
        the dataframe of offsets and lengths for each grib message and its
        attributes derived from an idx file
    raw_merged : bool
        Used for debugging to see all the columns in the merge. By default,
        it returns the index columns with the corrected time values plus
        the index metadata

    Returns
    -------

    pd.Dataframe : the index dataframe that will be used to read variable data from the grib file
    """

    idxdf = idxdf.reset_index().set_index("attrs")
    mapping = mapping.reset_index().set_index("attrs")
    mapping.drop(columns="uri", inplace=True)  # Drop the URI column from the mapping

    if not idxdf.index.is_unique:
        raise ValueError("Parsed idx data must have unique attrs to merge on!")

    if not mapping.index.is_unique:
        raise ValueError("Mapping data must have unique attrs to merge on!")

    # Merge the offset and length from the idx file with the varname, step and level from the mapping

    result = idxdf.merge(mapping, on="attrs", how="left", suffixes=("", "_mapping"))

    if raw_merged:
        return result
    else:
        # Get the grib_uri column from the idxdf and ignore the uri column from the mapping
        # We want the offset, length and uri of the index file with the varname, step and level of the mapping
        selected_results = result.rename(columns=dict(grib_uri="uri"))[
            [
                "varname",
                "typeOfLevel",
                "stepType",
                "name",
                "step",
                "level",
                "time",
                "valid_time",
                "uri",
                "offset",
                "length",
                "inline_value",
            ]
        ]
    # Drop the inline values from the mapping data
    selected_results.loc[:, "inline_value"] = None
    selected_results.loc[:, "time"] = run_time
    selected_results.loc[:, "valid_time"] = (
        selected_results.time + selected_results.step
    )
    logger.info("Dropping %d nan varnames", selected_results.varname.isna().sum())
    selected_results = selected_results.loc[~selected_results.varname.isna(), :]
    return selected_results.reset_index(drop=True)



import base64
import copy
import itertools
import fsspec.asyn
from typing import Any, cast
import warnings

import ujson

import fsspec.implementations.asyn_wrapper
import numpy as np
import zarr.storage


def dict_to_store(store_dict: dict):
    """Create an in memory zarr store backed by the given dictionary"""
    return zarr.storage.MemoryStore(read_only=False, store_dict=store_dict)


def refs_as_fs(
    refs,
    fs=None,
    remote_protocol=None,
    remote_options=None,
    asynchronous=True,
    **kwargs,
):
    """Convert a reference set to an fsspec filesystem"""
    if fs is not None:
        fs.asynchronous = asynchronous
    else:
        remote_options = remote_options.copy() if remote_options else {}
        remote_options["asynchronous"] = asynchronous
    fs = fsspec.filesystem(
        "reference",
        fo=refs,
        fs=fs,
        remote_protocol=remote_protocol,
        remote_options=remote_options,
        **kwargs,
        asynchronous=asynchronous,
    )
    return fs


def refs_as_store(
    refs, read_only=False, fs=None, remote_protocol=None, remote_options=None, **kwargs
):
    """Convert a reference set to a zarr store"""
    remote_options = remote_options or {}
    remote_options["asynchronous"] = True

    fss = refs_as_fs(
        refs,
        fs=fs,
        remote_protocol=remote_protocol,
        remote_options=remote_options,
        **kwargs,
    )
    return fs_as_store(fss, read_only=read_only)


def fs_as_store(fs: fsspec.asyn.AsyncFileSystem, read_only=False):
    """Open the refs as a zarr store

    Parameters
    ----------
    fs: fsspec.async.AsyncFileSystem
    read_only: bool

    Returns
    -------
    zarr.storage.Store or zarr.storage.Mapper, fsspec.AbstractFileSystem
    """
    if not fs.async_impl:
        try:
            from fsspec.implementations.asyn_wrapper import AsyncFileSystemWrapper

            fs = AsyncFileSystemWrapper(fs)
        except ImportError:
            raise ImportError(
                "Only fsspec>2024.10.0 supports the async filesystem wrapper "
                "required for working with reference filesystems. "
            )
    fs.asynchronous = True
    return zarr.storage.FsspecStore(fs, read_only=read_only)


def class_factory(func):
    """Experimental uniform API across function-based file scanners"""

    class FunctionWrapper:
        __doc__ = func.__doc__
        __module__ = func.__module__

        def __init__(self, url, storage_options=None, inline_threshold=100, **kwargs):
            self.url = url
            self.storage_options = storage_options
            self.inline = inline_threshold
            self.kwargs = kwargs

        def translate(self):
            return func(
                self.url,
                inline_threshold=self.inline,
                storage_options=self.storage_options,
                **self.kwargs,
            )

        def __str__(self):
            return f"<Single file to zarr processor using {func.__module__}.{func.__qualname__}>"

        __repr__ = __str__

    return FunctionWrapper


def consolidate(refs):
    """Turn raw references into output"""
    out = {}
    for k, v in refs.items():
        if hasattr(v, "to_bytes"):
            v = v.to_bytes()
        if isinstance(v, bytes):
            try:
                # easiest way to test if data is ascii
                out[k] = v.decode("ascii")
            except UnicodeDecodeError:
                out[k] = (b"base64:" + base64.b64encode(v)).decode()
        else:
            out[k] = v
    return {"version": 1, "refs": out}


def rename_target(refs, renames):
    """Utility to change URLs in a reference set in a predictable way

    For reference sets including templates, this is more easily done by
    using template overrides at access time; but rewriting the references
    and saving a new file means not having to do that every time.

    Parameters
    ----------
    refs: dict
        Reference set
    renames: dict[str, str]
        Mapping from the old URL (including protocol, if this is how they appear
        in the original) to new URL

    Returns
    -------
    dict: the altered reference set, which can be saved
    """
    fs = refs_as_fs(refs)  # to produce normalised refs
    refs = fs.references
    out = {}
    for k, v in refs.items():
        if isinstance(v, list) and v[0] in renames:
            out[k] = [renames[v[0]]] + v[1:]
        else:
            out[k] = v
    return consolidate(out)


def rename_target_files(
    url_in, renames, url_out=None, storage_options_in=None, storage_options_out=None
):
    """Perform URL renames on a reference set - read and write from JSON

    Parameters
    ----------
    url_in: str
        Original JSON reference set
    renames: dict
        URL renamings to perform (see ``renate_target``)
    url_out: str | None
        Where to write to. If None, overwrites original
    storage_options_in: dict | None
        passed to fsspec for opening url_in
    storage_options_out: dict | None
        passed to fsspec for opening url_out. If None, storage_options_in is used.

    Returns
    -------
    None
    """
    with fsspec.open(url_in, **(storage_options_in or {})) as f:
        old = ujson.load(f)
    new = rename_target(old, renames)
    if url_out is None:
        url_out = url_in
    if storage_options_out is None:
        storage_options_out = storage_options_in
    with fsspec.open(url_out, mode="wt", **(storage_options_out or {})) as f:
        ujson.dump(new, f)


def _encode_for_JSON(store):
    """Make store JSON encodable"""
    for k, v in store.copy().items():
        if isinstance(v, list):
            continue
        else:
            try:
                # minify JSON
                v = ujson.dumps(ujson.loads(v))
            except (ValueError, TypeError):
                pass
            try:
                store[k] = v.decode() if isinstance(v, bytes) else v
            except UnicodeDecodeError:
                store[k] = "base64:" + base64.b64encode(v).decode()
    return store


def encode_fill_value(v: Any, dtype: np.dtype, compressor: Any = None) -> Any:
    # early out
    if v is None:
        return v

    # Assure that v is a numpy array of the appropriate dtype
    v = np.asanyarray(v, dtype=dtype)
    # If v is 1D and only has 1 element, squeeze it to 0-D ndarray.
    # When v is not 0-D, calling int/float/bool on v will fail
    # We still need an ndarray in case dtype is a date
    v = v.squeeze()

    if dtype.kind == "V" and dtype.hasobject:
        if compressor is None:
            raise ValueError("missing compressor for object array")
        v = compressor.encode(v)
        v = str(base64.standard_b64encode(v), "ascii")
        return v
    if dtype.kind == "f":
        if np.isnan(v):
            return "NaN"
        elif np.isposinf(v):
            return "Infinity"
        elif np.isneginf(v):
            return "-Infinity"
        else:
            return float(v)
    elif dtype.kind in "ui":
        return int(v)
    elif dtype.kind == "b":
        return bool(v)
    elif dtype.kind in "c":
        c = cast(np.complex128, np.dtype(complex).type())
        v = (
            encode_fill_value(v.real, c.real.dtype, compressor),
            encode_fill_value(v.imag, c.imag.dtype, compressor),
        )
        return v
    elif dtype.kind in "SV":
        v = str(base64.standard_b64encode(v), "ascii")
        return v
    elif dtype.kind == "U":
        return v
    elif dtype.kind in "mM":
        return int(v.view("i8"))
    else:
        return v


def do_inline(store, threshold, remote_options=None, remote_protocol=None):
    """Replace short chunks with the value of that chunk and inline metadata

    The chunk may need encoding with base64 if not ascii, so actual
    length may be larger than threshold.
    """
    fs = refs_as_fs(
        store,
        remote_protocol=remote_protocol,
        remote_options=remote_options,
        asynchronous=False,
    )
    out = fs.references.copy()

    # Inlining is done when one of two conditions are satisfied:
    # 1. The item is small enough, i.e. smaller than the threshold specified in the function call
    # 2. The item is a metadata file, i.e. a .z* file
    get_keys = [
        k
        for k, v in out.items()
        if (isinstance(v, list) and len(v) == 3 and v[2] < threshold)
        or (
            isinstance(v, list)
            and len(v) == 1
            and isinstance(v[0], str)
            and v[0].split("/")[-1].startswith(".z")
        )
    ]
    values = fs.cat(get_keys)
    for k, v in values.items():
        try:
            # easiest way to test if data is ascii
            v.decode("ascii")
        except UnicodeDecodeError:
            v = b"base64:" + base64.b64encode(v)
        out[k] = v
    return out


def _inline_array(group, threshold, names, prefix=""):
    for name, thing in group.members():
        if prefix:
            prefix1 = f"{prefix}.{name}"
        else:
            prefix1 = name
        if isinstance(thing, zarr.Group):
            _inline_array(thing, threshold=threshold, prefix=prefix1, names=names)
        else:
            cond1 = threshold and thing.nbytes < threshold
            cond2 = prefix1 in names
            if cond1 or cond2:
                original_attrs = dict(thing.attrs)
                data = thing[...]
                arr = group.create_array(
                    name=name,
                    dtype=thing.dtype,
                    shape=thing.shape,
                    chunks=thing.shape,
                    fill_value=thing.fill_value,
                    overwrite=True,
                )
                arr[...] = data
                arr.attrs.update(original_attrs)


def inline_array(store, threshold=1000, names=None, remote_options=None):
    """Inline whole arrays by threshold or name, replace with a single metadata chunk

    Inlining whole arrays results in fewer keys. If the constituent keys were
    already inlined, this also results in a smaller file overall. No action is taken
    for arrays that are already of one chunk (they should be in

    Parameters
    ----------
    store: dict/JSON file
        reference set
    threshold: int
        Size in bytes below which to inline. Set to 0 to prevent inlining by size
    names: list[str] | None
        It the array name (as a dotted full path) appears in this list, it will
        be inlined irrespective of the threshold size. Useful for coordinates.
    remote_options: dict | None
        Needed to fetch data, if the required keys are not already individually inlined
        in the data.

    Returns
    -------
    amended references set (simple style)
    """
    fs = refs_as_fs(store, remote_options=remote_options or {})
    zarr_store = fs_as_store(fs, read_only=False)
    g = zarr.open_group(zarr_store, zarr_format=2)
    _inline_array(g, threshold, names=names or [])
    return fs.references


def subchunk(store, variable, factor):
    """
    Split uncompressed chunks into integer subchunks on the largest axis

    Parameters
    ----------
    store: dict
        reference set
    variable: str
        the named zarr variable (give as /-separated path if deep)
    factor: int
        the number of chunks each input chunk turns into. Must be an exact divisor
        of the original largest dimension length.

    Returns
    -------
    modified store
    """
    fs = fsspec.filesystem("reference", fo=store)
    store = fs.references
    meta_file = f"{variable}/.zarray"
    meta = ujson.loads(fs.cat(meta_file))
    if meta["compressor"] is not None:
        raise ValueError("Can only subchunk an uncompressed array")
    chunks_orig = meta["chunks"]
    chunk_new = []
    # plan
    multi = None
    for ind, this_chunk in enumerate(chunks_orig):
        if this_chunk == 1:
            chunk_new.append(1)
            continue
        elif this_chunk % factor == 0:
            chunk_new.extend([this_chunk // factor] + chunks_orig[ind + 1 :])
            break
        elif factor % this_chunk == 0:
            # if factor // chunks_orig[0] > 1:
            chunk_new.append(1)
            if multi is None:
                multi = this_chunk
            factor //= this_chunk
        else:
            raise ValueError("Must subchunk by exact integer factor")

    if multi:
        # TODO: this reloads the referenceFS; *maybe* reuses it
        return subchunk(store, variable, multi)

    # execute
    meta["chunks"] = chunk_new
    store = copy.deepcopy(store)
    store[meta_file] = ujson.dumps(meta)

    for k, v in store.copy().items():
        if k.startswith(f"{variable}/"):
            kpart = k[len(variable) + 1 :]
            if kpart.startswith(".z"):
                continue
            sep = "." if "." in kpart else "/"
            chunk_index = [int(_) for _ in kpart.split(sep)]
            if isinstance(v, (str, bytes)):
                # TODO: check this early, as some refs might have been edited already
                raise ValueError("Refusing to sub-chunk inlined data")
            if len(v) > 1:
                url, offset, size = v
            else:
                (url,) = v
                offset = 0
                size = fs.info(k)["size"]
            for subpart in range(factor):
                new_index = (
                    chunk_index[:ind]
                    + [chunk_index[ind] * factor + subpart]
                    + chunk_index[ind + 1 :]
                )
                newpart = sep.join(str(_) for _ in new_index)
                newv = [url, offset + subpart * size // factor, size // factor]
                store[f"{variable}/{newpart}"] = newv
    return store


def dereference_archives(references, remote_options=None):
    """Directly point to uncompressed byte ranges in ZIP/TAR archives

    If a set of references have been made for files contained within ZIP or
    (uncompressed) TAR archives, the "zip://..." and "tar://..." URLs should
    be converted to byte ranges in the overall file.

    Parameters
    ----------
    references: dict
        a simple reference set
    remote_options: dict or None
        For opening the archives
    """
    import zipfile
    import tarfile

    if "version" in references and references["version"] == 1:
        references = references["refs"]

    target_files = [l[0] for l in references.values() if isinstance(l, list)]
    target_files = {
        (t.split("::", 1)[1], t[:3])
        for t in target_files
        if t.startswith(("tar://", "zip://"))
    }

    # find all member file offsets in all archives
    offsets = {}
    for target, tar_or_zip in target_files:
        with fsspec.open(target, **(remote_options or {})) as tf:
            if tar_or_zip == "tar":
                tar = tarfile.TarFile(fileobj=tf)
                offsets[target] = {
                    ti.name: {"offset": ti.offset_data, "size": ti.size, "comp": False}
                    for ti in tar.getmembers()
                    if ti.isfile()
                }
            elif tar_or_zip == "zip":
                zf = zipfile.ZipFile(file=tf)
                offsets[target] = {}
                for zipinfo in zf.filelist:
                    if zipinfo.is_dir():
                        continue
                    # if uncompressed, include only the buffer. In compressed (DEFLATE), include
                    # also the header, and must use DeflateCodec
                    if zipinfo.compress_type == zipfile.ZIP_DEFLATED:
                        # TODO: find relevant .zarray and add filter directly
                        header = 0
                        warnings.warn(
                            "ZIP file contains compressed files, must use DeflateCodec"
                        )
                        tail = len(zipinfo.FileHeader())
                    elif zipinfo.compress_type == zipfile.ZIP_STORED:
                        header = len(zipinfo.FileHeader())
                        tail = 0
                    else:
                        comp = zipfile.compressor_names[zipinfo.compress_type]
                        raise ValueError(
                            f"ZIP compression method not supported: {comp}"
                        )
                    offsets[target][zipinfo.filename] = {
                        "offset": zipinfo.header_offset + header,
                        "size": zipinfo.compress_size + tail,
                        "comp": zipinfo.compress_type != zipfile.ZIP_STORED,
                    }

    # modify references
    mods = copy.deepcopy(references)
    for k, v in mods.items():
        if not isinstance(v, list):
            continue
        target = v[0].split("::", 1)[1]
        infile = v[0].split("::", 1)[0][6:]  # strip "zip://" or "tar://"
        if target not in offsets:
            continue
        detail = offsets[target][infile]
        if detail["comp"]:
            # leave compressed member file alone
            pass
        v[0] = target
        if len(v) == 1:
            v.append(detail["offset"])
            v.append(detail["size"])
        else:
            v[1] += detail["offset"]
    return mods


def _max_prefix(*strings):
    # https://stackoverflow.com/a/6719272/3821154
    def all_same(x):
        return all(x[0] == y for y in x)

    char_tuples = zip(*strings)
    prefix_tuples = itertools.takewhile(all_same, char_tuples)
    return "".join(x[0] for x in prefix_tuples)


def templateize(strings, min_length=10, template_name="u"):
    """Make prefix template for a set of strings

    Useful for condensing strings by extracting out a common prefix.
    If the common prefix is shorted than ``min_length``, the original
    strings are returned and the output templates are empty.

    Parameters
    ----------
    strings: List[str]
        inputs
    min_length: int
        Only perform transformm if the common prefix is at least this long.
    template_name: str
        The placeholder string, should be short.

    Returns
    -------
    templates: Dict[str, str], strings: List[str]
    Such that [s.format(**templates) for s in strings] recreates original strings list
    """
    prefix = _max_prefix(*strings)
    lpref = len(prefix)
    if lpref >= min_length:
        template = {template_name: prefix}
        strings = [("{%s}" % template_name) + s[lpref:] for s in strings]
    else:
        template = {}
    return template, strings


def translate_refs_serializable(refs: dict):
    """Translate a reference set to a serializable form, given that zarr
    v3 memory stores store data in buffers by default. This modifies the
    input dictionary in place, and returns a reference to it.

    It also fixes keys that have a leading slash, which is not appropriate for
    zarr v3 keys

    Parameters
    ----------
    refs: dict
        The reference set

    Returns
    -------
    dict
        A serializable form of the reference set
    """
    keys_to_remove = []
    new_keys = {}
    for k, v in refs.items():
        if isinstance(v, zarr.core.buffer.cpu.Buffer):
            key = k.removeprefix("/")
            new_keys[key] = v.to_bytes()
            keys_to_remove.append(k)
    for k in keys_to_remove:
        del refs[k]
    refs.update(new_keys)
    return refs



import ast
from dataclasses import dataclass
import io

import numcodecs
from numcodecs.abc import Codec
import numpy as np
import threading
import zlib
from zarr.core.array_spec import ArraySpec
from zarr.abc.codec import ArrayBytesCodec
from zarr.core.buffer import Buffer, NDBuffer
from zarr.core.common import JSON, parse_named_configuration
from zarr.registry import register_codec


class FillStringsCodec(Codec):
    """Sets fixed-length string fields to empty

    To be used with HDF fields of strings, to fill in the valules of the opaque
    16-byte string IDs.
    """

    codec_id = "fill_hdf_strings"

    def __init__(self, dtype, id_map=None):
        """
        Note: we must pass id_map using strings, because this is JSON-encoded
        by zarr.

        Parameters
        ----------
        id_map: None | str | dict(str, str)
        """
        if "[" in dtype:
            self.dtype = ast.literal_eval(dtype)
        else:
            self.dtype = dtype
        self.id_map = id_map

    def encode(self, buf):
        raise NotImplementedError

    def decode(self, buf, out=None):
        if isinstance(self.dtype, list):
            dt = [tuple(_) for _ in self.dtype]
        else:
            dt = self.dtype
        arr = np.frombuffer(buf, dtype=dt).copy()
        if arr.dtype.kind in "SU":
            if isinstance(self.id_map, dict):
                arr = np.array([self.id_map[_.decode()] for _ in arr], dtype="O")
            else:
                arr = np.full(arr.shape, self.id_map, dtype="O")
            return arr
        elif arr.dtype.kind == "V":
            dt2 = []
            for name in arr.dtype.names:
                if arr.dtype[name].kind in "SU":
                    dt2.append((name, "O"))
                else:
                    dt2.append((name, arr.dtype[name]))
            arr2 = np.empty(arr.shape, dtype=dt2)
            for name in arr.dtype.names:
                if arr[name].dtype.kind in "SU":
                    if isinstance(self.id_map, dict):
                        arr2[name][:] = [self.id_map[_.decode()] for _ in arr[name]]
                    else:
                        arr2[name][:] = self.id_map
                else:
                    arr2[name][:] = arr[name]

            return arr2


numcodecs.register_codec(FillStringsCodec, "fill_hdf_strings")


class GRIBCodec(numcodecs.abc.Codec):
    """
    Read GRIB stream of bytes as a message using eccodes
    """

    eclock = threading.RLock()

    codec_id = "grib"

    def __init__(self, var, dtype=None):
        self.var = var
        self.dtype = dtype

    def encode(self, buf):
        # on encode, pass through
        return buf

    def decode(self, buf, out=None):
        import eccodes

        if self.var in ["latitude", "longitude"]:
            var = self.var + "s"
            dt = self.dtype or "float64"
        else:
            var = "values"
            dt = self.dtype or "float32"
        with self.eclock:
            mid = eccodes.codes_new_from_message(bytes(buf))
            try:
                data = eccodes.codes_get_array(mid, var)
                missingValue = eccodes.codes_get_string(mid, "missingValue")
                if var == "values" and missingValue:
                    data[data == float(missingValue)] = np.nan
                if out is not None:
                    return numcodecs.compat.ndarray_copy(data, out)
                else:
                    return data.astype(dt, copy=False)

            finally:
                eccodes.codes_release(mid)


numcodecs.register_codec(GRIBCodec, "grib")


@dataclass(frozen=True)
class GRIBZarrCodec(ArrayBytesCodec):
    eclock = threading.RLock()

    var: str
    dtype: np.dtype

    def __init__(self, *, var: str, dtype: np.dtype) -> None:
        object.__setattr__(self, "var", var)
        object.__setattr__(self, "dtype", dtype)

    @classmethod
    def from_dict(cls, data: dict[str, JSON]) -> "GRIBZarrCodec":
        _, configuration_parsed = parse_named_configuration(
            data, "bytes", require_configuration=True
        )
        configuration_parsed = configuration_parsed or {}
        return cls(**configuration_parsed)  # type: ignore[arg-type]

    def to_dict(self) -> dict[str, JSON]:
        if self.endian is None:
            return {"name": "grib"}
        else:
            return {
                "name": "grib",
                "configuration": {"var": self.var, "dtype": self.dtype},
            }

    async def _decode_single(
        self,
        chunk_bytes: Buffer,
        chunk_spec: ArraySpec,
    ) -> NDBuffer:
        assert isinstance(chunk_bytes, Buffer)
        import eccodes

        if self.var in ["latitude", "longitude"]:
            var = self.var + "s"
            dt = self.dtype or "float64"
        else:
            var = "values"
            dt = self.dtype or "float32"

        with self.eclock:
            mid = eccodes.codes_new_from_message(chunk_bytes.to_bytes())
            try:
                data = eccodes.codes_get_array(mid, var)
                missingValue = eccodes.codes_get_string(mid, "missingValue")
                if var == "values" and missingValue:
                    data[data == float(missingValue)] = np.nan
                return data.astype(dt, copy=False)

            finally:
                eccodes.codes_release(mid)

    async def _encode_single(
        self,
        chunk_array: NDBuffer,
        chunk_spec: ArraySpec,
    ) -> Buffer | None:
        # This is a one way codec
        raise NotImplementedError

    def compute_encoded_size(
        self, input_byte_length: int, _chunk_spec: ArraySpec
    ) -> int:
        raise NotImplementedError


register_codec("grib", GRIBZarrCodec)


class AsciiTableCodec(numcodecs.abc.Codec):
    """Decodes ASCII-TABLE extensions in FITS files"""

    codec_id = "FITSAscii"

    def __init__(self, indtypes, outdtypes):
        """

        Parameters
        ----------
        indtypes: list[str]
            dtypes of the fields as in the table
        outdtypes: list[str]
            requested final dtypes
        """
        self.indtypes = indtypes
        self.outdtypes = outdtypes

    def decode(self, buf, out=None):
        indtypes = np.dtype([tuple(d) for d in self.indtypes])
        outdtypes = np.dtype([tuple(d) for d in self.outdtypes])
        arr = np.frombuffer(buf, dtype=indtypes)
        return arr.astype(outdtypes)

    def encode(self, _):
        pass


class VarArrCodec(numcodecs.abc.Codec):
    """Variable length arrays in a FITS BINTABLE extension"""

    codec_id = "FITSVarBintable"
    # https://heasarc.gsfc.nasa.gov/docs/software/fitsio/quick/node10.html
    ftypes = {"B": "uint8", "I": ">i2", "J": ">i4"}

    def __init__(self, dt_in, dt_out, nrow, types):
        self.dt_in = dt_in
        self.dt_out = dt_out
        self.nrow = nrow
        self.types = types

    def encode(self, _):
        raise NotImplementedError

    def decode(self, buf, out=None):
        dt_in = np.dtype(ast.literal_eval(self.dt_in))
        dt_out = np.dtype(ast.literal_eval(self.dt_out))
        arr = np.frombuffer(buf, dtype=dt_in, count=self.nrow)
        arr2 = np.empty((self.nrow,), dtype=dt_out)
        heap = buf[arr.nbytes :]
        for name in dt_out.names:
            if dt_out[name] == "O":
                dt = np.dtype(self.ftypes[self.types[name]])
                counts = arr[name][:, 0]
                offs = arr[name][:, 1]
                for i, (off, count) in enumerate(zip(offs, counts)):
                    arr2[name][i] = np.frombuffer(
                        heap[off : off + count * dt.itemsize], dtype=dt
                    )
            else:
                arr2[name][:] = arr[name][:]
        return arr2


numcodecs.register_codec(AsciiTableCodec, "FITSAscii")
numcodecs.register_codec(VarArrCodec, "FITSVarBintable")


class RecordArrayMember(Codec):
    """Read components of a record array (complex dtype)"""

    codec_id = "record_member"

    def __init__(self, member, dtype):
        """
        Parameters
        ----------
        member: str
            name of desired subarray
        dtype: list of lists
            description of the complex dtype of the overall record array. Must be both
            parsable by ``np.dtype()`` and also be JSON serialisable
        """
        self.member = member
        self.dtype = dtype

    def decode(self, buf, out=None):
        arr = np.frombuffer(buf, dtype=np.dtype(self.dtype))
        return arr[self.member].copy()

    def encode(self, buf):
        raise NotImplementedError


class DeflateCodec(Codec):
    """As implemented for members of zip-files

    The input buffer contains the file header as well as the compressed bytes
    """

    codec_id = "deflate"

    def decode(self, buf, out=None):
        import zipfile
        import struct

        head = buf[: zipfile.sizeFileHeader]
        *_, csize, usize, fnsize, extra_size = struct.unpack(
            zipfile.structFileHeader, head
        )

        zi = zipfile.ZipInfo()
        zi.compress_size = csize
        zi.file_size = usize
        zi.compress_type = zipfile.ZIP_DEFLATED

        b = io.BytesIO(buf)
        b.seek(zipfile.sizeFileHeader + fnsize + extra_size)
        zf = zipfile.ZipExtFile(b, mode="r", zipinfo=zi)
        return zf.read()

    def encode(self, buf):
        raise NotImplementedError


class ZlibCodec(Codec):
    codec_id = "zlib"

    def __init__(self):
        pass

    def decode(self, data, out=None):
        if out:
            out[:] = zlib.decompress(data)
            return out
        return zlib.decompress(data)

    def encode(self, buf):
        return zlib.compress(buf)



import asyncio
import collections.abc
import logging
import re
from typing import List
import warnings

import fsspec
import fsspec.utils
import numpy as np
import numcodecs
import ujson
import zarr
from zarr.core.buffer.core import default_buffer_prototype

from kerchunk.utils import consolidate, fs_as_store

logger = logging.getLogger("kerchunk.combine")


def drop(fields):
    """Generate example preprocessor removing given fields"""

    def preproc(refs):
        for k in list(refs):
            if k.startswith(fields):
                refs.pop(k)
        return refs

    return preproc


class MultiZarrToZarr:
    """
    Combine multiple kerchunk'd datasets into a single logical aggregate dataset

    :param path: str, list(str) or list(dict)
        Local paths, each containing a references JSON; or a list of references dicts.
        You may pass a list of reference dicts only, but then they will not have assicuated
        filenames; if you need filenames for producing coordinates, pass the list
        of filenames with path=, and the references with indicts=
    :param indicts: list(dict)
    :param concat_dims: str or list(str)
        Names of the dimensions to expand with
    :param coo_map: dict(str, selector)
        The special key "var" means the variable name in the output, which will be
        "VARNAME" by default (i.e., variable names are the same as in the input
        datasets). The default for any other coordinate is data:varname, i.e., look
        for an array with that name.

        Selectors ("how to get coordinate values from a dataset") can be:
            - a constant value (usually str for a var name, number for a coordinate)
            - a compiled regex ``re.Pattern``, which will be applied to the filename.
              Should return exactly one value
            - a string beginning "attr:" which will fetch this attribute from the zarr
              dataset of each path
            - a string beginning "vattr:{var}:" as above, but the attribute is taken from
              the array named var
            - "VARNAME" special value where a dataset contains multiple variables, just use
              the variable names as given
            - "INDEX" special value for the index of how far through the list of inputs
              we are so far
            - a string beginning "data:{var}" which will get the appropriate zarr array
              from each input dataset.
            - "cf:{var}", interpret the value of var using cftime, returning a datetime.
              These will be automatically re-encoded with cftime, *unless* you specify an
              "M8[*]" dtype for the coordinate, in which case a conversion will be attempted.
            - a list with the values that are known beforehand
            - a function with signature (index, fs, var, fn) -> value, where index is an int
              counter, fs is the file system made for the current input, var is the variable
              we are probing may be "var") and fn is the filename or None if dicts were used
              as input

    :param coo_dtypes: map(str, str|np.dtype)
        Coerce the final type of coordinate arrays (otherwise use numpy default)
    :param identical_dims: list[str]
        Variables that are to be copied across from the first input dataset, because they do not vary.
    :param target_options: dict
        Storage options for opening ``path``
    :param remote_protocol: str
        The protocol of the original data
    :param remote_options: dict
    :param inline_threshold: int
        Size below which binary blocks are included directly in the output
    :param preprocess: callable
        Acts on the references dict of all inputs before processing. See ``drop()``
        for an example.
    :param postprocess: callable
        Acts on the references dict before output.
        postprocess(dict)-> dict
    :param out: dict-like or None
        This allows you to supply an fsspec.implementations.reference.LazyReferenceMapper
        to write out parquet as the references get filled, or some other dictionary-like class
        to customise how references get stored
    :param append: bool
        If True, will load the references specified by out and add to them rather than starting
        from scratch. Assumes the same coordinates are being concatenated.
    """

    inline: int

    def __init__(
        self,
        path,
        indicts=None,
        coo_map=None,
        concat_dims=None,
        coo_dtypes=None,
        identical_dims=None,
        target_options=None,
        remote_protocol=None,
        remote_options=None,
        inline_threshold: int = 500,
        preprocess=None,
        postprocess=None,
        out=None,
    ):
        self._fss = None
        self._paths = None
        self._indicts = indicts
        self.ds = None
        self.path = path
        if concat_dims is None:
            self.concat_dims = list(coo_map)
        elif isinstance(concat_dims, str):
            self.concat_dims = [concat_dims]
        else:
            self.concat_dims = concat_dims
        self.coo_map = coo_map or {}
        self.coo_map.update(
            {
                c: "VARNAME" if c == "var" else f"data:{c}"
                for c in self.concat_dims
                if c not in self.coo_map
            }
        )
        logger.debug("Concat dims: %s", self.concat_dims)
        logger.debug("Coord map: %s", self.coo_map)
        self.coo_dtypes = coo_dtypes or {}
        self.target_options = target_options or {}
        self.remote_protocol = remote_protocol
        self.remote_options = remote_options or {}
        self.inline = inline_threshold
        self.cf_units = None
        self.identical_dims = identical_dims or []
        if set(self.coo_map).intersection(set(self.identical_dims)):
            raise ValueError("Values being mapped cannot also be identical")
        self.preprocess = preprocess
        self.postprocess = postprocess
        self.out = out if out is not None else {}
        self.coos = None
        self.done = set()

    @classmethod
    def append(
        cls,
        path,
        original_refs,
        remote_protocol=None,
        remote_options=None,
        target_options=None,
        **kwargs,
    ):
        """
        Update an existing combined reference set with new references

        There are two main usage patterns:

        - if the input ``original_refs`` is JSON, the combine happens in memory and the
          output should be written to JSON. This could then be optionally converted to parquet in a
          separate step
        - if ``original_refs`` is a lazy parquet reference set, then it will be amended in-place

        If you want to extend JSON references and output to parquet, you must first convert to
        parquet in the location you would like the final product to live.

        The other arguments should be the same as they were at the creation of the original combined
        reference set.

        NOTE: if the original combine used a postprocess function, it may be that this process
        functions, as the combine is done "before" postprocessing. Functions that only add information
        (as as setting attrs) would be OK.

        Parameters
        ----------
        path: list of reference sets to add. If remote/target options would be different
            to ``original_refs``, these can be as dicts or LazyReferenceMapper instances
        original_refs: combined reference set to be extended
        remote_protocol, remote_options, target_options: referring to ``original_refs```
        kwargs: to MultiZarrToZarr

        Returns
        -------
        MultiZarrToZarr
        """
        import xarray as xr

        storage_options = dict(
            fo=original_refs,
            remote_protocol=remote_protocol,
            remote_options=remote_options,
            target_options=target_options,
            asynchronous=True,
        )
        ds = xr.open_dataset(
            "reference://",
            engine="zarr",
            backend_kwargs={"consolidated": False},
            storage_options=storage_options,
        )
        z = zarr.open(
            "reference://",
            zarr_format=2,
            storage_options=storage_options,
            use_consolidated=False,
        )
        mzz = MultiZarrToZarr(
            path,
            out=z.store.fs.references,  # normalised dict or parquet/lazy
            remote_protocol=remote_protocol,
            remote_options=remote_options,
            target_options=target_options,
            **kwargs,
        )
        mzz.coos = {}
        for var, selector in mzz.coo_map.items():
            if (
                isinstance(selector, str)
                and selector.startswith("cf:")
                and "M" not in mzz.coo_dtypes.get(var, "")
            ):
                import cftime
                import datetime

                # undoing CF recoding in original input
                mzz.coos[var] = set()
                for c in ds[var].values:
                    value = cftime.date2num(
                        datetime.datetime.fromisoformat(str(c).split(".")[0]),
                        calendar=ds[var].attrs.get(
                            "calendar", ds[var].encoding.get("calendar", "standard")
                        ),
                        units=ds[var].attrs.get("units", ds[var].encoding["units"]),
                    )
                    value2 = cftime.num2date(
                        value,
                        calendar=ds[var].attrs.get(
                            "calendar", ds[var].encoding.get("calendar", "standard")
                        ),
                        units=ds[var].attrs.get("units", ds[var].encoding["units"]),
                    )
                    mzz.coos[var].add(value2)

            else:
                mzz.coos[var] = set(z[var][:])
        return mzz

    @property
    def fss(self):
        """filesystem instances being analysed, one per input dataset"""
        import collections.abc

        if self._fss is None:
            logger.debug("setup filesystems")
            if self._indicts is not None:
                fo_list = self._indicts
                self._paths = self.path
            elif isinstance(self.path[0], collections.abc.Mapping):
                fo_list = self.path
                self._paths = []
                for path in self.path:
                    self._paths.append(path.get("templates", {}).get("u", None))
            else:
                self._paths = []
                for of in fsspec.open_files(self.path, **self.target_options):
                    self._paths.append(of.full_name)
                fs = fsspec.core.url_to_fs(
                    self.path[0], asynchronous=True, **self.target_options
                )[0]
                try:
                    # JSON path
                    fo_list = fs.cat(self.path)
                    fo_list = [ujson.loads(v) for v in fo_list.values()]
                except (IOError, TypeError, ValueError):
                    # tries again sequentially in comprehension below
                    fo_list = self.path

            self._fss = [
                fsspec.filesystem(
                    "reference",
                    fo=fo,
                    remote_protocol=self.remote_protocol,
                    remote_options=self.remote_options,
                )
                for fo in fo_list
            ]
        return self._fss

    def _get_value(self, index, z, var, fn=None):
        """derive coordinate value(s) for given input dataset

        How to map from input to

        index: int
            Current place in the list of inputs
        z: zarr group
            Open for the current input
        var: str
            name of value to extract.
        fn: str
            filename
        """
        selector = self.coo_map[var]
        if isinstance(selector, collections.abc.Callable):
            o = selector(index, z, var, fn)
        elif isinstance(selector, list):
            o = selector[index]
        elif isinstance(selector, re.Pattern):
            regex_groups_o = selector.search(fn).groups()
            if len(regex_groups_o) == 0:
                o = selector.search(fn).group()
            else:
                o = regex_groups_o[0]
        elif not isinstance(selector, str):
            # constant, should be int or float
            o = selector
        elif selector == "VARNAME":
            # used for merging variable names across datasets
            o = [_ for _ in z if _ not in self.concat_dims + self.identical_dims]
            if len(o) > 1:
                raise ValueError(
                    "Multiple varnames found in dataset, please "
                    "provide a more specific selector"
                )
            o = o[0]
        elif selector == "INDEX":
            o = index
        elif selector.startswith("attr:"):
            o = z.attrs[selector.split(":", 1)[1]]
        elif selector.startswith("vattr:"):
            _, var, item = selector.split(":", 3)
            o = z[var].attrs[item]
        elif selector.startswith("data:"):
            o = z[selector.split(":", 1)[1]][...]
        elif selector.startswith("cf:"):
            import cftime

            datavar = z[selector.split(":", 1)[1]]
            o = datavar[...]
            units = datavar.attrs.get("units")
            calendar = datavar.attrs.get("calendar", "standard")
            o = cftime.num2date(o, units=units, calendar=calendar)
            if "M" in self.coo_dtypes.get(var, ""):
                o = np.array([_.isoformat() for _ in o], dtype=self.coo_dtypes[var])
            else:
                if self.cf_units is None:
                    self.cf_units = {}
                if var not in self.cf_units:
                    self.cf_units[var] = dict(units=units, calendar=calendar)
        else:
            o = selector  # must be a non-number constant - error?
        if var in self.coo_dtypes:
            o = np.array(o, dtype=self.coo_dtypes[var])
        logger.debug("Decode: %s -> %s", (selector, index, var, fn), o)
        return o

    async def _read_meta_files(self, m, files):
        """Helper to load multiple metadata files asynchronously"""
        res = {}
        for fn in files:
            exists = await m.exists(fn)
            if exists:
                content = await m.get(fn, prototype=default_buffer_prototype())
                res[fn] = ujson.dumps(ujson.loads(content.to_bytes()))
        return res

    def first_pass(self):
        """Accumulate the set of concat coords values across all inputs"""

        coos = self.coos or {c: set() for c in self.coo_map}
        for i, fs in enumerate(self.fss):
            if self.preprocess:
                self.preprocess(fs.references)
                # reset this to force references to update
                fs.dircache = None
                fs._dircache_from_items()

            logger.debug("First pass: %s", i)
            z_store = fs_as_store(fs, read_only=False)
            z = zarr.open_group(z_store, zarr_format=2, use_consolidated=False)
            for var in self.concat_dims:
                value = self._get_value(i, z, var, fn=self._paths[i])
                if isinstance(value, np.ndarray):
                    value = value.ravel()
                if isinstance(value, (np.ndarray, tuple, list)):
                    coos[var].update(value)
                else:
                    coos[var].add(value)

        self.coos = _reorganise(coos)
        for c, v in self.coos.items():
            if len(v) < 2:
                warnings.warn(
                    f"Concatenated coordinate '{c}' contains less than expected"
                    f"number of values across the datasets: {v}"
                )
        logger.debug("Created coordinates map")
        self.done.add(1)
        return coos

    def store_coords(self):
        """
        Write coordinate arrays into the output
        """
        kv = {}
        store = zarr.storage.MemoryStore(kv)
        group = zarr.open_group(store, zarr_format=2, use_consolidated=False)
        m = fs_as_store(self.fss[0], read_only=False)
        z = zarr.open(m, zarr_format=2, use_consolidated=False)
        for k, v in self.coos.items():
            if k == "var":
                # The names of the variables to write in the second pass, not a coordinate
                continue
            # parametrize the threshold value below?
            compressor = numcodecs.Zstd() if len(v) > 100 else None
            kw = {}
            if self.cf_units and k in self.cf_units:
                if "M" not in self.coo_dtypes.get(k, ""):
                    import cftime

                    data = cftime.date2num(v, **self.cf_units[k]).ravel()
                    kw["fill_value"] = 2**62

            elif all([isinstance(_, (tuple, list)) for _ in v]):
                v = sum([list(_) if isinstance(_, tuple) else _ for _ in v], [])
                data = np.array(v, dtype=self.coo_dtypes.get(k))
            else:
                data = np.concatenate(
                    [
                        np.atleast_1d(np.array(_, dtype=self.coo_dtypes.get(k)))
                        for _ in v
                    ]
                ).ravel()
            if "fill_value" not in kw:
                if data.dtype.kind == "i":
                    kw["fill_value"] = None
                elif k in z:
                    # Fall back to existing fill value
                    kw["fill_value"] = z[k].fill_value
            arr = group.create_array(
                name=k,
                shape=data.shape,
                overwrite=True,
                compressor=compressor,
                dtype=self.coo_dtypes.get(k, data.dtype),
                **kw,
            )
            arr[:] = data
            if k in z:
                # copy attributes if values came from an original variable
                arr.attrs.update(z[k].attrs)
            arr.attrs["_ARRAY_DIMENSIONS"] = [k]
            if self.cf_units and k in self.cf_units:
                if "M" in self.coo_dtypes.get(k, ""):
                    arr.attrs.pop("units", None)
                    arr.attrs.pop("calendar", None)
                else:
                    arr.attrs.update(self.cf_units[k])
            # TODO: rewrite .zarray/.zattrs with ujson to save space. Maybe make them by hand anyway.
        self.out.update(kv)
        logger.debug("Written coordinates")

        metadata = asyncio.run(self._read_meta_files(m, [".zgroup", ".zattrs"]))
        self.out.update(metadata)
        logger.debug("Written global metadata")
        self.done.add(2)

    def second_pass(self):
        """map every input chunk to the output"""
        # TODO: this stage cannot be rerun without clearing and rerunning store_coords too,
        #  because some code runs dependent on the current state of self.out
        chunk_sizes = {}  #
        skip = set()
        dont_skip = set()
        did_them = set()
        no_deps = None

        for i, fs in enumerate(self.fss):
            to_download = {}
            m = fs_as_store(fs, read_only=False)
            z = zarr.open(m, zarr_format=2, use_consolidated=False)

            if no_deps is None:
                # done first time only
                deps = [z[_].attrs.get("_ARRAY_DIMENSIONS", []) for _ in z]
                all_deps = set(sum(deps, []))
                no_deps = set(self.coo_map) - all_deps

            # Coordinate values for the whole of this dataset
            cvalues = {
                c: self._get_value(i, z, c, fn=self._paths[i]) for c in self.coo_map
            }
            var = cvalues.get("var", None)
            for c, cv in cvalues.copy().items():
                if isinstance(cv, np.ndarray):
                    cv = cv.ravel()
                if isinstance(cv, (np.ndarray, list, tuple)):
                    cv = tuple(sorted(set(cv)))[0]
                    cvalues[c] = cv

            dirs = fs.ls("", detail=False)
            while dirs:
                v = dirs.pop(0)
                if v in self.coo_map or v in skip or v.startswith(".z"):
                    # already made coordinate variables and metadata
                    continue
                fns = fs.ls(v, detail=False)
                if f"{v}/.zgroup" in fns:
                    # recurse into groups - copy meta, add to dirs to process and don't look
                    # for references in this dir
                    metadata = asyncio.run(
                        self._read_meta_files(m, [f"{v}/.zgroup", f"{v}/.zattrs"])
                    )
                    self.out.update(metadata)
                    dirs.extend([f for f in fns if not f.startswith(f"{v}/.z")])
                    continue
                if v in self.identical_dims:
                    if f"{v}/.zarray" in self.out:
                        continue
                    for k in fs.ls(v, detail=False):
                        if k.startswith(f"{v}/"):
                            self.out[k] = fs.references[k]
                    continue
                logger.debug("Second pass: %s, %s", i, v)

                zarray = asyncio.run(self._read_meta_files(m, [f"{v}/.zarray"]))[
                    f"{v}/.zarray"
                ]
                zarray = ujson.loads(zarray)
                if v not in chunk_sizes:
                    chunk_sizes[v] = zarray["chunks"]
                elif chunk_sizes[v] != zarray["chunks"]:
                    raise ValueError(
                        f"""Found chunk size mismatch:
                        at prefix {v} in iteration {i} (file {self._paths[i]})
                        new chunk: {chunk_sizes[v]}
                        chunks so far: {zarray["chunks"]}"""
                    )
                chunks = chunk_sizes[v]
                zattr_meta = asyncio.run(self._read_meta_files(m, [f"{v}/.zattrs"]))
                zattrs = ujson.loads(zattr_meta.get(f"{v}/.zattrs", "{}"))
                coords = zattrs.get("_ARRAY_DIMENSIONS", [])
                if zarray["shape"] and not coords:
                    coords = list("ikjlm")[: len(zarray["shape"])]

                if v not in dont_skip and v in all_deps:
                    # this is an input coordinate
                    # a coordinate is any array appearing in its own or other array's _ARRAY_DIMENSIONS
                    skip.add(v)
                    for k in fs.ls(v, detail=False):
                        if k.rsplit("/", 1)[-1].startswith(".z"):
                            self.out[k] = fs.cat(k)
                        else:
                            self.out[k] = fs.references[k]
                    continue

                dont_skip.add(v)  # don't check for coord or identical again

                coord_order = [
                    c for c in self.concat_dims if c not in coords and c != "var"
                ] + coords

                # create output array, accounting for shape, chunks and dim dependencies
                if (var or v) not in did_them:
                    did_them.add(var or v)
                    shape = []
                    ch = []
                    for c in coord_order:
                        if c in self.coos:
                            shape.append(
                                self.coos[c].size
                                if isinstance(self.coos[c], np.ndarray)
                                else len(self.coos[c])
                            )
                        else:
                            shape.append(zarray["shape"][coords.index(c)])
                        ch.append(chunks[coords.index(c)] if c in coords else 1)

                    zarray["shape"] = shape
                    zarray["chunks"] = ch
                    zattrs["_ARRAY_DIMENSIONS"] = coord_order
                    self.out[f"{var or v}/.zarray"] = ujson.dumps(zarray)
                    # other attributes copied as-is from first occurrence of this array
                    self.out[f"{var or v}/.zattrs"] = ujson.dumps(zattrs)
                else:
                    k = self.out[f"{var or v}/.zarray"]
                    ch = ujson.loads(k)["chunks"]

                for fn in fns:
                    # loop over the chunks and copy the references
                    if ".z" in fn:
                        continue
                    key_parts = fn.split("/")[-1].split(".")
                    key = f"{var or v}/"
                    for loc, c in enumerate(coord_order):
                        if c in self.coos:
                            cv = cvalues[c]
                            ind = np.searchsorted(self.coos[c], cv)
                            if c in coords:
                                key += str(
                                    ind // ch[loc] + int(key_parts[coords.index(c)])
                                )
                            else:
                                key += str(ind // ch[loc])
                        else:
                            key += key_parts[coords.index(c)]
                        key += "."
                    key = key.rstrip(".")

                    ref = fs.references.get(fn)
                    if (
                        self.inline > 0
                        and isinstance(ref, list)
                        and (
                            (len(ref) > 1 and ref[2] < self.inline)
                            or fs.info(fn)["size"] < self.inline
                        )
                    ):
                        to_download[key] = fn
                    else:
                        self.out[key] = fs.references[fn]
            if to_download:
                bits = fs.cat(list(to_download.values()))
                for key, fn in to_download.items():
                    self.out[key] = bits[fn]
        self.done.add(3)

    def translate(self, filename=None, storage_options=None):
        """Perform all stages and return the resultant references dict

        If filename and storage options are given, the output is written to this
        file using ujson and fsspec.
        """
        if 1 not in self.done:
            self.first_pass()
        if 2 not in self.done:
            self.store_coords()
        if 3 not in self.done:
            self.second_pass()
        if 4 not in self.done:
            if self.postprocess is not None:
                self.out = self.postprocess(self.out)
            self.done.add(4)
        if isinstance(self.out, dict):
            out = consolidate(self.out)
        else:
            self.out.flush()
            out = self.out
        if filename is not None:
            with fsspec.open(filename, mode="wt", **(storage_options or {})) as f:
                ujson.dump(out, f)
        return out


def _reorganise(coos):
    # reorganise and sort coordinate values
    # extracted here to enable testing
    out = {}
    for k, arr in coos.items():
        out[k] = np.array(sorted(arr))
    return out


def merge_vars(files, storage_options=None):
    """Merge variables across datasets with identical coordinates

    :param files: list(dict), list(str) or list(fsspec.OpenFile)
        List of reference dictionaries or list of paths to reference json files to be merged
    :param storage_options: dict
        Dictionary containing kwargs to `fsspec.open_files`
    """
    if isinstance(files[0], collections.abc.Mapping):
        fo_list = files
        merged = fo_list[0].copy()
        for file in fo_list[1:]:
            refs = file["refs"]
            merged["refs"].update(refs)
    else:
        fo_list = fsspec.open_files(files, mode="rb", **(storage_options or {}))
        with fo_list[0] as f:
            merged = ujson.load(f)
        for file in fo_list[1:]:
            with file as f:
                refs = ujson.load(f)["refs"]
            merged["refs"].update(refs)
    return merged


def concatenate_arrays(
    files,
    storage_options=None,
    axis=0,
    key_seperator=".",
    path=None,
    check_arrays=False,
):
    """Simple concatenate of one zarr array along an axis

    Assumes that each array is identical in shape/type.

    If the inputs are groups, provide the path to the contained array, and all other arrays
    will be ignored. You could concatentate the arrays separately and then recombine them
    with ``merge_vars``.

    Parameters
    ----------
    files: list[dict] | list[str]
        Input reference sets, maybe generated by ``kerchunk.zarr.single_zarr``
    storage_options: dict | None
        To create the filesystems, such at target/remote protocol and target/remote options
    key_seperator: str
        "." or "/", how the zarr keys are stored
    path: str or None
        If the datasets are groups rather than simple arrays, this is the location in the
        group hierarchy to concatenate. The group structure will be recreated.
    check_arrays: bool
        Whether we check the size and chunking of the inputs. If True, and an
        inconsistency is found, an exception is raised. If False (default), the
        user is expected to be certain that the chunking and shapes are
        compatible.
    """
    out = {}
    if path is None:
        path = ""
    else:
        path = "/".join(path.rstrip(".").rstrip("/").split(".")) + "/"

    def _replace(l: list, i: int, v) -> list:
        l = l.copy()
        l[i] = v
        return l

    n_files = len(files)

    chunks_offset = 0
    for i, fn in enumerate(files):
        fs = fsspec.filesystem("reference", fo=fn, **(storage_options or {}))
        zarray = ujson.load(fs.open(f"{path}.zarray"))
        shape = zarray["shape"]
        chunks = zarray["chunks"]
        n_chunks, rem = divmod(shape[axis], chunks[axis])
        n_chunks += rem > 0

        if i == 0:
            base_shape = _replace(shape, axis, None)
            base_chunks = chunks
            # result_* are modified in-place
            result_zarray = zarray
            result_shape = shape
            for name in [".zgroup", ".zattrs", f"{path}.zattrs"]:
                if name in fs.references:
                    out[name] = fs.references[name]
        else:
            result_shape[axis] += shape[axis]

        # Safety checks
        if check_arrays:
            if _replace(shape, axis, None) != base_shape:
                expected_shape = (
                    f"[{', '.join(map(str, _replace(base_shape, axis, '*')))}]"
                )
                raise ValueError(
                    f"Incompatible array shape at index {i}. Expected {expected_shape}, got {shape}."
                )
            if chunks != base_chunks:
                raise ValueError(
                    f"Incompatible array chunks at index {i}. Expected {base_chunks}, got {chunks}."
                )
            if i < (n_files - 1) and rem != 0:
                raise ValueError(
                    f"Array at index {i} has irregular chunking at its boundary. "
                    "This is only allowed for the final array."
                )

        # Referencing the offset chunks
        for key in fs.find(""):
            if key.startswith(f"{path}.z") or not key.startswith(path):
                continue
            parts = key.lstrip(path).split(key_seperator)
            parts[axis] = str(int(parts[axis]) + chunks_offset)
            key2 = path + key_seperator.join(parts)
            out[key2] = fs.references[key]

        chunks_offset += n_chunks

    out[f"{path}.zarray"] = ujson.dumps(result_zarray)

    return consolidate(out)


def auto_dask(
    urls: List[str],
    single_driver: type,
    single_kwargs: dict,
    mzz_kwargs: dict,
    n_batches: int,
    remote_protocol=None,
    remote_options=None,
    filename=None,
    output_options=None,
):
    """Batched tree combine using dask.

    If you wish to run on a distributed cluster (recommended), create
    a client before calling this function.

    Parameters
    ----------
    urls: list[str]
        input dataset URLs
    single_driver: class
        class with ``translate()`` method
    single_kwargs: to pass to single-input driver
    mzz_kwargs: passed to ``MultiZarrToZarr`` for each batch
    n_batches: int
        Number of MZZ instances in the first combine stage. Maybe set equal
        to the number of dask workers, or a multple thereof.
    remote_protocol: str | None
    remote_options: dict
        To fsspec for opening the remote files
    filename: str | None
        Ouput filename, if writing
    output_options
        If ``filename`` is not None, open it with these options

    Returns
    -------
    reference set
    """
    import dask

    # make delayed functions
    single_task = dask.delayed(lambda x: single_driver(x, **single_kwargs).translate())
    post = mzz_kwargs.pop("postprocess", None)
    inline = mzz_kwargs.pop("inline_threshold", None)
    # TODO: if single files produce list of reference sets (e.g., grib2)
    batch_task = dask.delayed(
        lambda u, x: MultiZarrToZarr(
            u,
            indicts=x,
            remote_protocol=remote_protocol,
            remote_options=remote_options,
            **mzz_kwargs,
        ).translate()
    )

    # sort out kwargs
    dims = mzz_kwargs.get("concat_dims", [])
    dims += [k for k in mzz_kwargs.get("coo_map", []) if k not in dims]
    kwargs = {"concat_dims": dims}
    if post:
        kwargs["postprocess"] = post
    if inline:
        kwargs["inline_threshold"] = inline
    for field in ["remote_protocol", "remote_options", "coo_dtypes", "identical_dims"]:
        if field in mzz_kwargs:
            kwargs[field] = mzz_kwargs[field]
    final_task = dask.delayed(
        lambda x: MultiZarrToZarr(
            x, remote_options=remote_options, remote_protocol=remote_protocol, **kwargs
        ).translate(filename, output_options)
    )

    # make delayed calls
    tasks = [single_task(u) for u in urls]
    tasks_per_batch = -(-len(tasks) // n_batches)
    tasks2 = []
    for batch in range(n_batches):
        in_tasks = tasks[batch * tasks_per_batch : (batch + 1) * tasks_per_batch]
        u = urls[batch * tasks_per_batch : (batch + 1) * tasks_per_batch]
        if in_tasks:
            # skip if on last iteration and no remaining tasks
            tasks2.append(batch_task(u, in_tasks))
    return dask.compute(final_task(tasks2))[0]


class JustLoad:
    """For auto_dask, in the case that single file references already exist"""

    def __init__(self, url, storage_options=None):
        self.url = url
        self.storage_options = storage_options or {}

    def translate(self):
        with fsspec.open(self.url, mode="rt", **self.storage_options) as f:
            return ujson.load(f)



"""
# Extract, Load and Build Kerchunk aggregations on the fly from hierarchical metadata and kerchunk indexes.

MIT License Copyright (c) 2023 Camus Energy

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import copy
import re
import unittest
import logging
import gzip
import os
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from kerchunk.grib2 import (
    scan_grib,
    grib_tree,
    correct_hrrr_subhf_step,
    extract_datatree_chunk_index,
    strip_datavar_chunks,
    reinflate_grib_store,
    AggregationType,
    build_idx_grib_mapping,
    parse_grib_idx,
    map_from_index,
    read_store,
    write_store,
)
from kerchunk.utils import refs_as_store
import fsspec
import zarr
import ujson
import tempfile
import typing
import io

# https://github.com/pydata/xarray/issues/9984
# until datatree/xarray supports zarr3
pytest.skip(allow_module_level=True)
logger = logging.getLogger(__name__)

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class DataExtractorTests(unittest.TestCase):
    def test_integration(self):
        # Small test file...
        uri = f"{THIS_DIR}/grib_idx_fixtures/hrrr.wrfsubhf.sample.grib2"

        scanned_msg_groups = scan_grib(uri)
        corrected_msg_groups = [
            correct_hrrr_subhf_step(msg) for msg in scanned_msg_groups
        ]
        grib_tree_store = grib_tree(corrected_msg_groups)
        zg = zarr.open_group("reference://", storage_options={"fo": grib_tree_store})
        self.assertIsInstance(zg["refc/instant/atmosphere/refc"], zarr.Array)
        self.assertIsInstance(zg["vbdsf/avg/surface/vbdsf"], zarr.Array)
        self.assertEqual(
            zg["vbdsf/avg/surface"].attrs["coordinates"],
            "surface latitude longitude step time valid_time",
        )
        self.assertEqual(
            zg["refc/instant/atmosphere"].attrs["coordinates"],
            "atmosphere latitude longitude step time valid_time",
        )
        # Assert that the fill value is set correctly
        self.assertIs(zg["refc"]["instant"]["atmosphere"]["step"].fill_value, np.nan)

        np.testing.assert_array_equal(
            zg["refc"]["instant"]["atmosphere"]["time"][:], np.array([1665709200])
        )

        # Read it with data tree and assert the same...
        dt = xr.open_datatree(
            "reference://",
            storage_options={"fo": grib_tree_store},
            engine="zarr",
            consolidated=False,
        )
        # Assert a few things... but if it loads we are mostly done.
        np.testing.assert_array_equal(
            dt["refc"]["instant"]["atmosphere"]["time"].values[:],
            np.array([np.datetime64("2022-10-14T01:00:00")]),
        )
        self.assertDictEqual(
            dt["refc"].attrs, dict(name="Maximum/Composite radar reflectivity")
        )

        # Now try the extract and reinflate methods
        self.assertEqual(len(grib_tree_store["refs"]), 55)
        deflated_grib_tree = copy.deepcopy(grib_tree_store)
        kind = extract_datatree_chunk_index(dt, deflated_grib_tree, grib=True)
        strip_datavar_chunks(deflated_grib_tree)
        self.assertEqual(len(deflated_grib_tree["refs"]), 45)

        axes = [
            pd.Index(
                [
                    pd.timedelta_range(
                        start="0 minutes",
                        end="60 minutes",
                        freq="15min",
                        closed="left",
                        name="001 hour",
                    ),
                ],
                name="step",
            ),
            pd.date_range(
                "2022-10-14T00:00:00",
                "2022-10-14T02:00:00",
                freq="15min",
                name="valid_time",
            ),
        ]

        zstore = reinflate_grib_store(
            axes=axes,
            aggregation_type=AggregationType.HORIZON,
            chunk_index=kind,
            zarr_ref_store=deflated_grib_tree,
        )
        # Back to the same number of keys!
        self.assertEqual(len(zstore["refs"]), 55)

        dt = xr.open_datatree(
            "reference://",
            storage_options={"fo": zstore},
            engine="zarr",
            consolidated=False,
        )
        for node in dt.subtree:
            if node.has_data:
                with self.subTest(node_path=node.path):
                    np.testing.assert_array_equal(
                        node.time.values, node.valid_time.values - node.step.values
                    )

                    npath = node.path.lstrip("/")

                    # Assert the values that should be nonnull based on where the chunk is in the time slice
                    expected_null = np.full((1, 9, 1059, 1799), True)
                    expected_null[0, 5, :, :] = False
                    for var in node.data_vars.values():
                        vpath = f"{npath}/{var.name}"
                        with self.subTest(var_name=var.name):
                            key_set = sorted(
                                [
                                    key
                                    for key in zstore["refs"].keys()
                                    if key.startswith(vpath)
                                ]
                            )
                            expected_keys = [
                                f"{vpath}/.zarray",
                                f"{vpath}/.zattrs",
                                f"{vpath}/0.5.0.0",
                            ]
                            self.assertListEqual(key_set, expected_keys)

                            np.testing.assert_array_equal(
                                np.isnan(var.values[:, :, :, :]), expected_null
                            )

    def test_build_idx_grib_mapping(self):
        """
        This test builds a mapping between idx and grib metadata from one runtime and applies it in another, asserting the
        mapped index matches the direct index (fixture from the next test).
        :return:
        """
        grib_uri = os.path.join(THIS_DIR, "grib_idx_fixtures")

        # Define pairs of files with the same horizon but from different runtimes to validate idx mapping
        datasets = {
            "hrrr.wrfsubhf": [
                {
                    "20221014": "hrrr.t01z.wrfsubhf00.grib2.test-limit-10",
                    "20231104": "hrrr.t01z.wrfsubhf00.grib2.test-limit-10",
                },
                {
                    "20221014": "hrrr.t03z.wrfsubhf09.grib2.test-limit-10",
                    "20231104": "hrrr.t01z.wrfsubhf09.grib2.test-limit-10",
                },
            ],
            "gfs.pgrb2.0p25": [
                {
                    "20221014": "gfs.t00z.pgrb2.0p25.f000.test-limit-10",
                    "20231104": "gfs.t00z.pgrb2.0p25.f000.test-limit-10",
                },
                {
                    "20221014": "gfs.t06z.pgrb2.0p25.f021.test-limit-10",
                    "20231104": "gfs.t00z.pgrb2.0p25.f021.test-limit-10",
                },
            ],
            "hrrr.wrfsfcf": [
                {
                    "20221014": "hrrr.t01z.wrfsfcf00.grib2.test-limit-10",
                    "20231104": "hrrr.t01z.wrfsfcf00.grib2.test-limit-10",
                },
                {
                    "20221014": "hrrr.t09z.wrfsfcf05.grib2.test-limit-10",
                    "20231104": "hrrr.t01z.wrfsfcf05.grib2.test-limit-10",
                },
            ],
        }
        for sample_prefix, input_pairs in datasets.items():
            for fnames in input_pairs:
                # First, build a mapping from the 2022 10 14 data
                mapping_fname = fnames["20221014"]
                basename = os.path.join(grib_uri, "20221014", mapping_fname)
                with self.subTest(sample_prefix=sample_prefix):
                    mapping = build_idx_grib_mapping(
                        basename=basename,
                        mapper=correct_hrrr_subhf_step,
                    )

                    mapping = mapping.assign(
                        uri=lambda x: x.uri.str.replace(THIS_DIR, ""),
                        grib_uri=lambda x: x.uri.str.replace(THIS_DIR, ""),
                        idx_uri=lambda x: x.uri.str.replace(THIS_DIR, ""),
                    )

                    # # To update the test fixture
                    # write_path = os.path.join(
                    #     THIS_DIR,
                    #     "grib_idx_fixtures",
                    #     sample_prefix,
                    #     "20221014",
                    #     f"{mapping_fname}.idx_grib_mapping.parquet",
                    # )
                    # mapping.to_parquet(write_path)

                    test_path = os.path.join(
                        THIS_DIR,
                        "grib_idx_fixtures",
                        sample_prefix,
                        "20221014",
                        f"{mapping_fname}.idx_grib_mapping.parquet",
                    )
                    # Build the mapping from idx to cfgrib metadata and assert it matches the fixture
                    expected = pd.read_parquet(test_path, engine="fastparquet")
                    pd.testing.assert_frame_equal(mapping, expected)

                    # parse the idx files for 20231104 and compare the mapped result to the direct indexed result
                    test_name = fnames["20231104"]
                    basename = os.path.join(grib_uri, "20231104", test_name)

                    idxdf = parse_grib_idx(
                        basename=basename,
                    )

                    # Get the runtime hour from the filename as we would in prod
                    matcher = re.compile(r"\w*\.t(?P<hour>\d{2})z\..*")
                    matched = matcher.match(test_name)
                    if not matched:
                        raise ValueError(f"test name {test_name} hour not matched")
                    runtime_hour = matched.groupdict()["hour"]

                    # hard code the runtime date for now and build the index using the
                    # 2022-10-14 mapping for 2023-11-04 idx file!
                    mapped_index = map_from_index(
                        pd.Timestamp(f"2023-11-04T{runtime_hour}"), mapping, idxdf
                    )

                    mapped_index = mapped_index.assign(
                        uri=lambda x: x.uri.str.replace(THIS_DIR, "")
                    )

                    # Read the expected fixture - created by test_kerchunk_indexing
                    kindex_test_path = os.path.join(
                        THIS_DIR,
                        "grib_idx_fixtures",
                        sample_prefix,
                        "20231104",
                        f"{test_name}.kindex.parquet",
                    )
                    expected = pd.read_parquet(kindex_test_path, engine="fastparquet")

                    expected = expected.set_index(
                        ["varname", "typeOfLevel", "stepType", "step", "level"]
                    ).sort_index()
                    mapped_index = mapped_index.set_index(
                        ["varname", "typeOfLevel", "stepType", "step", "level"]
                    ).sort_index()

                    pd.testing.assert_index_equal(expected.index, mapped_index.index)

                    pd.testing.assert_frame_equal(
                        expected[["time", "valid_time"]],
                        mapped_index[["time", "valid_time"]],
                    )

                    # The grib index can build inline values so only compare the offset, length, and uri where
                    # the inline value is not present

                    expected_subset = expected.loc[
                        expected.inline_value.isna(), ["offset", "length", "uri"]
                    ]

                    mapped_index_subset = mapped_index.loc[
                        expected.inline_value.isna(), ["offset", "length", "uri"]
                    ]
                    pd.testing.assert_frame_equal(expected_subset, mapped_index_subset)

    def test_kerchunk_indexing(self):
        """
        This test builds the grib metadata index for a set of forecasts and asserts it has not changed from what is
        recorded in the fixture files
        :return:
        """
        TEST_DATE = "20231104"
        grib_uri = os.path.join(THIS_DIR, "grib_idx_fixtures", TEST_DATE)

        datasets = {
            "hrrr.wrfsubhf": [
                "hrrr.t01z.wrfsubhf00.grib2.test-limit-10",
                "hrrr.t01z.wrfsubhf09.grib2.test-limit-10",
            ],
            "gfs.pgrb2.0p25": [
                "gfs.t00z.pgrb2.0p25.f000.test-limit-10",
                "gfs.t00z.pgrb2.0p25.f021.test-limit-10",
            ],
            "hrrr.wrfsfcf": [
                "hrrr.t01z.wrfsfcf00.grib2.test-limit-10",
                "hrrr.t01z.wrfsfcf05.grib2.test-limit-10",
            ],
        }
        for sample_prefix, fnames in datasets.items():
            for fname in fnames:
                with self.subTest(sample_prefix=sample_prefix, fname=fname):

                    basename = os.path.join(grib_uri, fname)
                    if "hrrr.wrfsubhf" == sample_prefix:
                        grib_tree_store = grib_tree(
                            [
                                correct_hrrr_subhf_step(scan)
                                for scan in scan_grib(basename)
                            ]
                        )
                    else:
                        grib_tree_store = grib_tree(scan_grib(basename))

                    dt = xr.open_datatree(
                        "reference://",
                        storage_options={"fo": grib_tree_store},
                        engine="zarr",
                        consolidated=False,
                    )

                    kindex = extract_datatree_chunk_index(
                        dt, grib_tree_store, grib=True
                    )

                    # Strip the current working directory
                    kindex = kindex.assign(
                        uri=lambda x: x.uri.str.replace(THIS_DIR, "")
                    )

                    # # To update the test fixture
                    # write_path = os.path.join(
                    #     THIS_DIR,
                    #     "grib_idx_fixtures",
                    #     sample_prefix,
                    #     TEST_DATE,
                    #     f"{fname}.kindex.parquet",
                    # )
                    # kindex.to_parquet(write_path)

                    test_path = os.path.join(
                        THIS_DIR,
                        "grib_idx_fixtures",
                        sample_prefix,
                        TEST_DATE,
                        f"{fname}.kindex.parquet",
                    )
                    expected = pd.read_parquet(test_path, engine="fastparquet")
                    pd.testing.assert_frame_equal(kindex, expected)

    def test_index_extraction(self):
        for sample_prefix in ["hrrr.wrfsubhf", "hrrr.wrfsfcf", "gfs.pgrb2.0p25"]:
            with self.subTest(sample_prefix):
                self._test_index_extraction(sample_prefix)

    def _read_sample_prefix(self, sample_prefix: str) -> tuple[xr.DataTree, dict]:
        read_path = os.path.join(
            THIS_DIR, "grib_idx_fixtures", sample_prefix, "subset.json.gz"
        )
        with gzip.open(read_path, "rt") as f:
            scanned_msgs = ujson.load(typing.cast(io.TextIOWrapper, f))

        if "subhf" in sample_prefix:
            scanned_msgs = [correct_hrrr_subhf_step(msg) for msg in scanned_msgs]
        grib_tree_store = grib_tree(scanned_msgs)

        dt = xr.open_datatree(
            "reference://",
            storage_options={"fo": grib_tree_store},
            engine="zarr",
            consolidated=False,
        )
        return dt, grib_tree_store

    def _test_index_extraction(self, sample_prefix: str):
        dt, grib_tree_store = self._read_sample_prefix(sample_prefix)

        k_index = extract_datatree_chunk_index(dt, grib_tree_store, grib=True)

        # # To update the test fixture
        # write_path = os.path.join(
        #     THIS_DIR, "grib_idx_fixtures", sample_prefix, "kerchunk_index.parquet"
        # )
        # k_index.to_parquet(write_path)

        test_path = os.path.join(
            THIS_DIR, "grib_idx_fixtures", sample_prefix, "kerchunk_index.parquet"
        )
        expected = pd.read_parquet(test_path, engine="fastparquet")
        pd.testing.assert_frame_equal(k_index, expected)

    def test_strip_datavar_chunks(self):
        for sample_prefix, pre, post in [
            ("hrrr.wrfsubhf", 104, 65),
            ("hrrr.wrfsfcf", 108, 63),
            ("gfs.pgrb2.0p25", 917, 205),
        ]:
            with self.subTest(sample_prefix=sample_prefix, pre=pre, post=post):
                self._test_strip_datavar_chunks(sample_prefix, pre, post)

    def _test_strip_datavar_chunks(self, sample_prefix: str, pre: int, post: int):
        dt, grib_tree_store = self._read_sample_prefix(sample_prefix)

        self.assertEqual(len(grib_tree_store["refs"]), pre)
        strip_datavar_chunks(grib_tree_store)
        self.assertEqual(len(grib_tree_store["refs"]), post)

        # # To update the test grib_idx_fixtures
        # write_path = os.path.join(
        #     TESTS_DIR, "grib_idx_fixtures", sample_prefix
        # )
        # dynamic_zarr_store.write_store(write_path, grib_tree_store)

        test_path = os.path.join(THIS_DIR, "grib_idx_fixtures", sample_prefix)
        expected = read_store(test_path)
        self.assertDictEqual(grib_tree_store, expected)

    def test_read_write_store(self):
        data = {
            "version": 1,
            "refs": {
                ".zgroup": '{"zarr_format":2}',
                ".zattrs": '{"GRIB_centre":"kwbc","GRIB_centreDescription":"US National Weather Service - NCEP ","GRIB_edition":2,"GRIB_subCentre":0,"coordinates":"heightAboveGround latitude longitude step time valid_time","institution":"US National Weather Service - NCEP "}',
                "u/.zarray": '{"chunks":[1059,1799],"compressor":null,"dtype":"<f8","fill_value":null,"filters":[{"dtype":"float64","id":"grib","var":"u"}],"order":"C","shape":[1059,1799],"zarr_format":2}',
                "u/0.0": ["{{u}}", 3653893, 1088857],
                "u/.zattrs": '{"GRIB_DxInMetres":3000.0,"GRIB_DyInMetres":3000.0,"GRIB_LaDInDegrees":38.5,"GRIB_Latin1InDegrees":38.5,"GRIB_Latin2InDegrees":38.5,"GRIB_LoVInDegrees":262.5,"GRIB_NV":0,"GRIB_Nx":1799,"GRIB_Ny":1059,"GRIB_cfName":"eastward_wind","GRIB_cfVarName":"u","GRIB_dataType":"fc","GRIB_gridDefinitionDescription":"Lambert Conformal can be secant or tangent, conical or bipolar","GRIB_gridType":"lambert","GRIB_iScansNegatively":0,"GRIB_jPointsAreConsecutive":0,"GRIB_jScansPositively":1,"GRIB_latitudeOfFirstGridPointInDegrees":21.138123,"GRIB_latitudeOfSouthernPoleInDegrees":0.0,"GRIB_longitudeOfFirstGridPointInDegrees":237.280472,"GRIB_longitudeOfSouthernPoleInDegrees":0.0,"GRIB_missingValue":3.4028234663852886e+38,"GRIB_name":"U component of wind","GRIB_numberOfPoints":1905141,"GRIB_paramId":131,"GRIB_shortName":"u","GRIB_stepType":"instant","GRIB_stepUnits":1,"GRIB_typeOfLevel":"heightAboveGround","GRIB_units":"m s**-1","_ARRAY_DIMENSIONS":["y","x"],"long_name":"U component of wind","standard_name":"eastward_wind","units":"m s**-1"}',
                "heightAboveGround/.zarray": '{"chunks":[],"compressor":null,"dtype":"<f8","fill_value":null,"filters":null,"order":"C","shape":[],"zarr_format":2}',
                "heightAboveGround/0": "\x00\x00\x00\x00\x00\x00T@",
                "heightAboveGround/.zattrs": '{"_ARRAY_DIMENSIONS":[],"long_name":"height above the surface","positive":"up","standard_name":"height","units":"m"}',
                "latitude/.zarray": '{"chunks":[1059,1799],"compressor":null,"dtype":"<f8","fill_value":null,"filters":[{"dtype":"float64","id":"grib","var":"latitude"}],"order":"C","shape":[1059,1799],"zarr_format":2}',
                "latitude/0.0": ["{{u}}", 3653893, 1088857],
                "latitude/.zattrs": '{"_ARRAY_DIMENSIONS":["y","x"],"long_name":"latitude","standard_name":"latitude","units":"degrees_north"}',
                "longitude/.zarray": '{"chunks":[1059,1799],"compressor":null,"dtype":"<f8","fill_value":null,"filters":[{"dtype":"float64","id":"grib","var":"longitude"}],"order":"C","shape":[1059,1799],"zarr_format":2}',
                "longitude/0.0": ["{{u}}", 3653893, 1088857],
                "longitude/.zattrs": '{"_ARRAY_DIMENSIONS":["y","x"],"long_name":"longitude","standard_name":"longitude","units":"degrees_east"}',
                "step/.zarray": '{"chunks":[],"compressor":null,"dtype":"<f8","fill_value":null,"filters":null,"order":"C","shape":[],"zarr_format":2}',
                "step/0": "\x00\x00\x00\x00\x00\x00\x00\x00",
                "step/.zattrs": '{"_ARRAY_DIMENSIONS":[],"long_name":"time since forecast_reference_time","standard_name":"forecast_period","units":"hours"}',
                "time/.zarray": '{"chunks":[],"compressor":null,"dtype":"<i8","fill_value":null,"filters":null,"order":"C","shape":[],"zarr_format":2}',
                "time/0": "base64:ENAUZQAAAAA=",
                "time/.zattrs": '{"_ARRAY_DIMENSIONS":[],"calendar":"proleptic_gregorian","long_name":"initial time of forecast","standard_name":"forecast_reference_time","units":"seconds since 1970-01-01T00:00:00"}',
                "valid_time/.zarray": '{"chunks":[],"compressor":null,"dtype":"<i8","fill_value":null,"filters":null,"order":"C","shape":[],"zarr_format":2}',
                "valid_time/0": "base64:ENAUZQAAAAA=",
                "valid_time/.zattrs": '{"_ARRAY_DIMENSIONS":[],"calendar":"proleptic_gregorian","long_name":"time","standard_name":"time","units":"seconds since 1970-01-01T00:00:00"}',
            },
            "templates": {"u": "testdata/hrrr.t01z.wrfsubhf00.grib2"},
        }
        with tempfile.TemporaryDirectory(suffix=".test") as ntd:
            write_store(ntd, data)

            result = read_store(ntd)
        self.assertDictEqual(data, result)

    def _reinflate_grib_store_dataset(self):
        datasets = [
            "hrrr.wrfsfcf",
            "gfs.pgrb2.0p25",
            "hrrr.wrfsubhf",
        ]
        for name in datasets:
            yield name

    def _reinflate_grib_store_aggregation(self):
        # Provide some general axes - this will subset the dataset for the selected times
        aggregations = {
            AggregationType.HORIZON: [
                pd.Index(
                    [
                        pd.timedelta_range(
                            start="0 minutes",
                            end="60 minutes",
                            freq="60min",
                            closed="left",
                            name="000 hour",
                        ),
                        pd.timedelta_range(
                            start="60 minutes",
                            end="120 minutes",
                            freq="60min",
                            closed="left",
                            name="001 hour",
                        ),
                    ],
                    name="step",
                ),
                pd.date_range(
                    "2023-09-28T00:00",
                    "2023-09-28T03:00",
                    freq="60min",
                    name="valid_time",
                ),
            ],
            AggregationType.VALID_TIME: [
                pd.timedelta_range("0 min", "300 min", freq="60 min", name="step"),
                pd.DatetimeIndex(
                    ["2023-09-28T02:00", "2023-09-28T04:00"], name="valid_time"
                ),
            ],
            AggregationType.RUN_TIME: [
                pd.timedelta_range("0 min", "120 min", freq="15 min", name="step"),
                pd.DatetimeIndex(["2023-09-28T00:00", "2023-09-28T02:00"], name="time"),
            ],
            AggregationType.BEST_AVAILABLE: [
                pd.date_range(
                    "2023-09-28T00:00",
                    "2023-09-28T10:00",
                    freq="60min",
                    name="valid_time",
                ),
                pd.DatetimeIndex(["2023-09-28T02:00"], name="time"),
            ],
        }

        for aggregation, axes in aggregations.items():
            yield aggregation, axes

    def _reinflate_grib_store(
        self,
        dataset: str,
        aggregation: AggregationType,
        axes: list[pd.Index],
    ):
        kind = pd.read_parquet(
            os.path.join(
                THIS_DIR, "grib_idx_fixtures", dataset, "test_reinflate.parquet"
            ),
            engine="fastparquet",
        )

        zstore = reinflate_grib_store(
            axes=axes,
            aggregation_type=aggregation,
            chunk_index=kind,
            zarr_ref_store=read_store(
                os.path.join(THIS_DIR, "grib_idx_fixtures", dataset)
            ),
        )
        dt = xr.open_datatree(refs_as_store(zstore), engine="zarr")
        for node in dt.subtree:
            if not node.has_data:
                continue
            with self.subTest(node_path=node.path):

                match aggregation:
                    case AggregationType.HORIZON | AggregationType.BEST_AVAILABLE:

                        self.assertEqual(node.time.dims, node.valid_time.dims)
                        self.assertEqual(node.time.dims, node.step.dims)
                        np.testing.assert_array_equal(
                            node.time.values, node.valid_time.values - node.step.values
                        )
                    case AggregationType.VALID_TIME:
                        steps2d = np.tile(
                            node.step.values, (node.valid_time.shape[0], 1)
                        )
                        valid_times2d = np.tile(
                            np.reshape(node.valid_time.values, (-1, 1)),
                            (1, node.step.shape[0]),
                        )

                        np.testing.assert_array_equal(
                            node.time.values, valid_times2d - steps2d
                        )
                    case AggregationType.RUN_TIME:
                        steps2d = np.tile(node.step.values, (node.time.shape[0], 1))
                        times2d = np.tile(
                            np.reshape(node.time.values, (-1, 1)),
                            (1, node.step.shape[0]),
                        )
                        np.testing.assert_array_equal(
                            times2d, node.valid_time.values - steps2d
                        )
                    case _:
                        raise RuntimeError("uhoh - unknown aggregation!")

                npath = node.path.lstrip("/")
                # Can't read nonlocal data here, but we can assert the correct keys are present
                for var in node.data_vars.values():
                    vpath = f"{npath}/{var.name}"
                    with self.subTest(var_name=var.name):
                        key_set = sorted(
                            [
                                key
                                for key in zstore["refs"].keys()
                                if key.startswith(vpath)
                            ]
                        )

                        # # To update test grib_idx_fixtures
                        # write_path = os.path.join(THIS_DIR, "grib_idx_fixtures", dataset, "reinflate", aggregation.value, f"{vpath}_chunks.json")
                        # with fsspec.open(write_path, "w",) as f:
                        #     f.write(ujson.dumps(key_set, indent=2))

                        test_path = os.path.join(
                            THIS_DIR,
                            "grib_idx_fixtures",
                            dataset,
                            "reinflate",
                            aggregation.value,
                            f"{vpath}_chunks.json",
                        )
                        with fsspec.open(test_path, "r") as f:
                            expected_keys = ujson.loads(f.read())

                        self.assertListEqual(key_set, expected_keys)

    def test_reinflate_grib_store(self):
        for dataset in self._reinflate_grib_store_dataset():
            for aggregation, axes in self._reinflate_grib_store_aggregation():
                with self.subTest(dataset=dataset, aggregation=aggregation):
                    self._reinflate_grib_store(dataset, aggregation, axes)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    unittest.main()



import pytest

import fsspec
import numpy as np
import zarr

import kerchunk.combine
import kerchunk.zarr
import kerchunk.df
from kerchunk.utils import fs_as_store, refs_as_store


@pytest.mark.parametrize(
    "arrays,chunks,axis",
    [
        (
            [np.arange(10), np.arange(10, 20)],
            (2,),
            0,
        ),
        (
            [np.arange(12), np.arange(12, 36), np.arange(36, 42)],
            (6,),
            0,
        ),
        (
            # Terminal chunk does not need to be filled
            [np.arange(5), np.arange(5, 10), np.arange(10, 17)],
            (5,),
            0,
        ),
        (
            [
                np.broadcast_to(np.arange(6), (10, 6)),
                np.broadcast_to(np.arange(7, 10), (10, 3)),
            ],
            (10, 3),
            1,
        ),
        (
            [
                np.broadcast_to(np.arange(6), (10, 6)).T,
                np.broadcast_to(np.arange(7, 10), (10, 3)).T,
            ],
            (3, 10),
            0,
        ),
    ],
)
def test_success(tmpdir, arrays, chunks, axis, m):
    fns = []
    refs = []
    for i, x in enumerate(arrays):
        fn = f"{tmpdir}/out{i}.zarr"
        g = zarr.open(fn, zarr_format=2)
        arr = g.create_array("x", shape=x.shape, dtype=x.dtype, chunks=chunks)
        arr[:] = x
        fns.append(fn)
        ref = kerchunk.zarr.single_zarr(fn, inline=0)
        refs.append(ref)

    out = kerchunk.combine.concatenate_arrays(
        refs, axis=axis, path="x", check_arrays=True
    )

    store = refs_as_store(out)
    g = zarr.open(store, zarr_format=2)
    assert (g["x"][:] == np.concatenate(arrays, axis=axis)).all()

    try:
        import fastparquet
    except ImportError:
        return
    kerchunk.df.refs_to_dataframe(out, "memory://out.parq")
    storage_options = dict(
        fo="memory://out.parq",
        remote_protocol="file",
        skip_instance_cache=True,
    )
    g = zarr.open("reference://", zarr_format=2, storage_options=storage_options)
    assert (g["x"][:] == np.concatenate(arrays, axis=axis)).all()

    kerchunk.df.refs_to_dataframe(out, "memory://out.parq", record_size=1)
    storage_options = dict(
        fo="memory://out.parq",
        remote_protocol="file",
        skip_instance_cache=True,
    )
    g = zarr.open("reference://", zarr_format=2, storage_options=storage_options)
    assert (g["x"][:] == np.concatenate(arrays, axis=axis)).all()


def test_fail_chunks(tmpdir):
    fn1 = f"{tmpdir}/out1.zarr"
    fn2 = f"{tmpdir}/out2.zarr"
    x1 = np.arange(10)
    x2 = np.arange(10, 20)
    g = zarr.open(fn1, zarr_format=2)
    arr = g.create_array("x", shape=x1.shape, dtype=x1.dtype, chunks=(2,))
    arr[:] = x1
    g = zarr.open(fn2, zarr_format=2)
    arr = g.create_array("x", shape=x2.shape, dtype=x2.dtype, chunks=(3,))
    arr[:] = x2

    ref1 = kerchunk.zarr.single_zarr(fn1, inline=0)
    ref2 = kerchunk.zarr.single_zarr(fn2, inline=0)

    with pytest.raises(ValueError, match=r"Incompatible array chunks at index 1.*"):
        kerchunk.combine.concatenate_arrays([ref1, ref2], path="x", check_arrays=True)


def test_fail_shape(tmpdir):
    fn1 = f"{tmpdir}/out1.zarr"
    fn2 = f"{tmpdir}/out2.zarr"
    x1 = np.arange(12).reshape(6, 2)
    x2 = np.arange(12, 24)
    g = zarr.open(fn1, zarr_format=2)
    arr = g.create_array("x", shape=x1.shape, dtype=x1.dtype)
    arr[:] = x1
    g = zarr.open(fn2, zarr_format=2)
    arr = g.create_array("x", shape=x2.shape, dtype=x2.dtype)
    arr[:] = x2

    ref1 = kerchunk.zarr.single_zarr(fn1, inline=0)
    ref2 = kerchunk.zarr.single_zarr(fn2, inline=0)

    with pytest.raises(ValueError, match=r"Incompatible array shape at index 1.*"):
        kerchunk.combine.concatenate_arrays([ref1, ref2], path="x", check_arrays=True)


def test_fail_irregular_chunk_boundaries(tmpdir):
    fn1 = f"{tmpdir}/out1.zarr"
    fn2 = f"{tmpdir}/out2.zarr"
    x1 = np.arange(10)
    x2 = np.arange(10, 24)
    g = zarr.open(fn1, zarr_format=2)
    arr = g.create_array("x", shape=x1.shape, dtype=x1.dtype, chunks=(4,))
    arr[:] = x1
    g = zarr.open(fn2, zarr_format=2)
    arr = g.create_array("x", shape=x2.shape, dtype=x2.dtype, chunks=(4,))
    arr[:] = x2

    ref1 = kerchunk.zarr.single_zarr(fn1, inline=0)
    ref2 = kerchunk.zarr.single_zarr(fn2, inline=0)

    with pytest.raises(ValueError, match=r"Array at index 0 has irregular chunking.*"):
        kerchunk.combine.concatenate_arrays([ref1, ref2], path="x", check_arrays=True)



import base64
import copy
import io
import logging
from collections import defaultdict
from typing import Iterable, List, Dict, Set
import ujson

import fsspec
import zarr
import xarray
import numpy as np

from kerchunk.utils import (
    class_factory,
    _encode_for_JSON,
    dict_to_store,
    fs_as_store,
    translate_refs_serializable,
)
from kerchunk.codecs import GRIBCodec
from kerchunk.combine import MultiZarrToZarr, drop
from kerchunk._grib_idx import (
    parse_grib_idx,
    build_idx_grib_mapping,
    map_from_index,
    strip_datavar_chunks,
    reinflate_grib_store,
    extract_datatree_chunk_index,
    AggregationType,
    read_store,
    write_store,
)


try:
    import cfgrib
except ModuleNotFoundError as err:  # pragma: no cover
    if err.name == "cfgrib":
        raise ImportError(
            "cfgrib is needed to kerchunk GRIB2 files. Please install it with "
            "`conda install -c conda-forge cfgrib`. See https://github.com/ecmwf/cfgrib "
            "for more details."
        )


# cfgrib copies over certain GRIB attributes
# but renames them to CF-compliant values
ATTRS_TO_COPY_OVER = {
    "long_name": "GRIB_name",
    "units": "GRIB_units",
    "standard_name": "GRIB_cfName",
}

logger = logging.getLogger("grib2-to-zarr")


def _split_file(f: io.FileIO, skip=0):
    if hasattr(f, "size"):
        size = f.size
    else:
        size = f.seek(0, 2)
        f.seek(0)
    part = 0

    while f.tell() < size:
        logger.debug(f"extract part {part + 1}")
        head = f.read(1024)
        if b"GRIB" not in head:
            f.seek(-4, 1)
            continue
        ind = head.index(b"GRIB")
        start = f.tell() - len(head) + ind
        part_size = int.from_bytes(head[ind + 12 : ind + 16], "big")
        f.seek(start)
        yield start, part_size, f.read(part_size)
        part += 1
        if skip and part >= skip:
            break


def _store_array(store, z, data, var, inline_threshold, offset, size, attr):
    nbytes = data.dtype.itemsize
    for i in data.shape:
        nbytes *= i

    shape = tuple(data.shape or ())
    if nbytes < inline_threshold:
        logger.debug(f"Store {var} inline")
        d = z.create_array(
            name=var,
            shape=shape,
            chunks=shape,
            dtype=data.dtype,
            fill_value=attr.get("missingValue", None),
            compressor=None,
        )
        if hasattr(data, "tobytes"):
            b = data.tobytes()
        else:
            b = data.build_array().tobytes()
        try:
            # easiest way to test if data is ascii
            b.decode("ascii")
        except UnicodeDecodeError:
            b = b"base64:" + base64.b64encode(data)
        store[f"{var}/0"] = b.decode("ascii")
    else:
        logger.debug(f"Store {var} reference")
        d = z.create_array(
            name=var,
            shape=shape,
            chunks=shape,
            dtype=data.dtype,
            fill_value=attr.get("missingValue", None),
            filters=[GRIBCodec(var=var, dtype=str(data.dtype))],
            compressor=None,
        )
        store[f"{var}/" + ".".join(["0"] * len(shape))] = ["{{u}}", offset, size]
    d.attrs.update(attr)


def contains_valid_level(message_keys: Set) -> bool:
    """Check if the given set of message_keys contain a valid level value.
    Some types of level, like depthBelowLandLayer for GEFS grib files,
    represent slices of levels by "topLevel" and "bottomLevel" rather
    than a discrete level value described by "level".
    see https://github.com/fsspec/kerchunk/issues/559

    Args:
        message_keys: Set of keys to evaluate

    Returns:
        True if message_keys contain a valid level value, False otherwise
    """
    return "level" in message_keys or "topLevel" in message_keys


def scan_grib(
    url,
    common=None,
    storage_options=None,
    inline_threshold=100,
    skip=0,
    filter={},
) -> List[Dict]:
    """
    Generate references for a GRIB2 file

    Parameters
    ----------

    url: str
        File location
    common_vars: (depr, do not use)
    storage_options: dict
        For accessing the data, passed to filesystem
    inline_threshold: int
        If given, store array data smaller than this value directly in the output
    skip: int
        If non-zero, stop processing the file after this many messages
    filter: dict
        keyword filtering. For each key, only messages where the key exists and has
        the exact value or is in the given set, are processed.
        E.g., the cf-style filter ``{'typeOfLevel': 'heightAboveGround', 'level': 2}``
        only keeps messages where heightAboveGround==2.

    Returns
    -------

    list(dict): references dicts in Version 1 format, one per message in the file
    """
    import eccodes

    storage_options = storage_options or {}
    logger.debug(f"Open {url}")

    # This is hardcoded a lot in cfgrib!
    # valid_time is added if "time" and "step" are present in time_dims
    # These are present by default
    # TIME_DIMS = ["step", "time", "valid_time"]

    out = []
    with fsspec.open(url, "rb", **storage_options) as f:
        logger.debug(f"File {url}")
        for offset, size, data in _split_file(f, skip=skip):
            store_dict = {}
            store = dict_to_store(store_dict)

            mid = eccodes.codes_new_from_message(data)
            m = cfgrib.cfmessage.CfMessage(mid)

            # It would be nice to just have a list of valid keys
            # There does not seem to be a nice API for this
            # 1. message_grib_keys returns keys coded in the message
            # 2. There exist "computed" keys, that are functions applied on the data
            # 3. There are also aliases!
            #    e.g. "number" is an alias of "perturbationNumber", and cfgrib uses this alias
            # So we stick to checking membership in 'm', which ends up doing
            # a lot of reads.
            message_keys = set(m.message_grib_keys())
            # The choices here copy cfgrib :(
            # message_keys.update(cfgrib.dataset.INDEX_KEYS)
            # message_keys.update(TIME_DIMS)
            # print("totalNumber" in cfgrib.dataset.INDEX_KEYS)
            # Adding computed keys adds a lot that isn't added by cfgrib
            # message_keys.extend(m.computed_keys)

            shape = (m["Ny"], m["Nx"])
            # thank you, gribscan
            native_type = eccodes.codes_get_native_type(m.codes_id, "values")
            data_size = eccodes.codes_get_size(m.codes_id, "values")
            coordinates = []

            good = True
            for k, v in (filter or {}).items():
                if k not in m:
                    good = False
                elif isinstance(v, (list, tuple, set)):
                    if m[k] not in v:
                        good = False
                elif m[k] != v:
                    good = False
            if good is False:
                continue

            z = zarr.open_group(store, zarr_format=2)
            global_attrs = {
                f"GRIB_{k}": m[k]
                for k in cfgrib.dataset.GLOBAL_ATTRIBUTES_KEYS
                if k in m
            }
            if "GRIB_centreDescription" in global_attrs:
                # follow CF compliant renaming from cfgrib
                global_attrs["institution"] = global_attrs["GRIB_centreDescription"]
            z.attrs.update(global_attrs)

            if data_size < inline_threshold:
                # read the data
                vals = m["values"].reshape(shape)
            else:
                # dummy array to match the required interface
                vals = np.empty(shape, dtype=native_type)
                assert vals.size == data_size

            attrs = {
                # Follow cfgrib convention and rename key
                f"GRIB_{k}": m[k]
                for k in cfgrib.dataset.DATA_ATTRIBUTES_KEYS
                + cfgrib.dataset.EXTRA_DATA_ATTRIBUTES_KEYS
                + cfgrib.dataset.GRID_TYPE_MAP.get(m["gridType"], [])
                if k in m
            }
            for k, v in ATTRS_TO_COPY_OVER.items():
                if v in attrs:
                    attrs[k] = attrs[v]

            # try to use cfVarName if available,
            # otherwise use the grib shortName
            varName = m["cfVarName"]
            if varName in ("undef", "unknown"):
                varName = m["shortName"]
            _store_array(
                store_dict, z, vals, varName, inline_threshold, offset, size, attrs
            )
            if "typeOfLevel" in message_keys and contains_valid_level(message_keys):
                name = m["typeOfLevel"]
                coordinates.append(name)
                # convert to numpy scalar, so that .tobytes can be used for inlining
                # dtype=float is hardcoded in cfgrib
                data = np.array(m["level"], dtype=float)[()]
                try:
                    attrs = cfgrib.dataset.COORD_ATTRS[name]
                except KeyError:
                    logger.debug(f"Couldn't find coord {name} in dataset")
                    attrs = {}
                attrs["_ARRAY_DIMENSIONS"] = []
                _store_array(
                    store_dict, z, data, name, inline_threshold, offset, size, attrs
                )
            dims = (
                ["y", "x"]
                if m["gridType"] in cfgrib.dataset.GRID_TYPES_2D_NON_DIMENSION_COORDS
                else ["latitude", "longitude"]
            )
            z[varName].attrs["_ARRAY_DIMENSIONS"] = dims

            for coord in cfgrib.dataset.COORD_ATTRS:
                coord2 = {
                    "latitude": "latitudes",
                    "longitude": "longitudes",
                    "step": "step:int",
                }.get(coord, coord)
                try:
                    x = m.get(coord2)
                except eccodes.WrongStepUnitError as e:
                    logger.warning(
                        "Ignoring coordinate '%s' for varname '%s', raises: eccodes.WrongStepUnitError(%s)",
                        coord2,
                        varName,
                        e,
                    )
                    continue

                if x is None:
                    continue
                coordinates.append(coord)
                inline_extra = 0
                if isinstance(x, np.ndarray) and x.size == data_size:
                    if (
                        m["gridType"]
                        in cfgrib.dataset.GRID_TYPES_2D_NON_DIMENSION_COORDS
                    ):
                        dims = ["y", "x"]
                        x = x.reshape(vals.shape)
                    else:
                        dims = [coord]
                        if coord == "latitude":
                            x = x.reshape(vals.shape)[:, 0].copy()
                        elif coord == "longitude":
                            x = x.reshape(vals.shape)[0].copy()
                        # force inlining of x/y/latitude/longitude coordinates.
                        # since these are derived from analytic formulae
                        # and are not stored in the message
                        inline_extra = x.nbytes + 1
                elif np.isscalar(x):
                    # convert python scalars to numpy scalar
                    # so that .tobytes can be used for inlining
                    x = np.array(x)[()]
                    dims = []
                else:
                    x = np.array([x])
                    dims = [coord]
                attrs = cfgrib.dataset.COORD_ATTRS[coord]
                _store_array(
                    store_dict,
                    z,
                    x,
                    coord,
                    inline_threshold + inline_extra,
                    offset,
                    size,
                    attrs,
                )
                z[coord].attrs["_ARRAY_DIMENSIONS"] = dims
            if coordinates:
                z.attrs["coordinates"] = " ".join(coordinates)

            translate_refs_serializable(store_dict)
            out.append(
                {
                    "version": 1,
                    "refs": _encode_for_JSON(store_dict),
                    "templates": {"u": url},
                }
            )
    logger.debug("Done")
    return out


GribToZarr = class_factory(scan_grib)


def example_combine(
    filter={"typeOfLevel": "heightAboveGround", "level": 2}
):  # pragma: no cover
    """Create combined dataset of weather measurements at 2m height

    Ten consecutive timepoints from ten 120MB files on s3.
    Example usage:

    >>> tot = example_combine()
    >>> ds = xr.open_dataset("reference://", engine="zarr", backend_kwargs={
    ...        "consolidated": False,
    ...        "storage_options": {"fo": tot, "remote_options": {"anon": True}}})
    """
    files = [
        "s3://noaa-hrrr-bdp-pds/hrrr.20190101/conus/hrrr.t22z.wrfsfcf01.grib2",
        "s3://noaa-hrrr-bdp-pds/hrrr.20190101/conus/hrrr.t23z.wrfsfcf01.grib2",
        "s3://noaa-hrrr-bdp-pds/hrrr.20190102/conus/hrrr.t00z.wrfsfcf01.grib2",
        "s3://noaa-hrrr-bdp-pds/hrrr.20190102/conus/hrrr.t01z.wrfsfcf01.grib2",
        "s3://noaa-hrrr-bdp-pds/hrrr.20190102/conus/hrrr.t02z.wrfsfcf01.grib2",
        "s3://noaa-hrrr-bdp-pds/hrrr.20190102/conus/hrrr.t03z.wrfsfcf01.grib2",
        "s3://noaa-hrrr-bdp-pds/hrrr.20190102/conus/hrrr.t04z.wrfsfcf01.grib2",
        "s3://noaa-hrrr-bdp-pds/hrrr.20190102/conus/hrrr.t05z.wrfsfcf01.grib2",
        "s3://noaa-hrrr-bdp-pds/hrrr.20190102/conus/hrrr.t06z.wrfsfcf01.grib2",
    ]
    so = {"anon": True, "default_cache_type": "readahead"}

    out = [scan_grib(u, storage_options=so, filter=filter) for u in files]
    out = sum(out, [])
    mzz = MultiZarrToZarr(
        out,
        remote_protocol="s3",
        preprocess=drop(("valid_time", "step")),
        remote_options=so,
        concat_dims=["time", "var"],
        identical_dims=["heightAboveGround", "latitude", "longitude"],
    )
    return mzz.translate()


def grib_tree(
    message_groups: Iterable[Dict],
    remote_options=None,
) -> Dict:
    """
    Build a hierarchical data model from a set of scanned grib messages.

    The iterable input groups should be a collection of results from scan_grib. Multiple grib files can
    be processed together to produce an FMRC like collection.
    The time (reference_time) and step coordinates will be used as concat_dims in the MultiZarrToZarr
    aggregation. Each variable name will become a group with nested subgroups representing the grib
    step type and grib level. The resulting hierarchy can be opened as a zarr_group or a xarray datatree.
    Grib message variable names that decode as "unknown" are dropped
    Grib typeOfLevel attributes that decode as unknown are treated as a single group
    Grib steps that are missing due to WrongStepUnitError are patched with NaT
    The input message_groups should not be modified by this method

    Parameters
    ----------
    message_groups: iterable[dict]
        a collection of zarr store like dictionaries as produced by scan_grib
    remote_options: dict
        remote options to pass to MultiZarrToZarr

    Returns
    -------
    dict: A new zarr store like dictionary for use as a reference filesystem mapper with zarr
    or xarray datatree
    """
    # Hard code the filters in the correct order for the group hierarchy
    filters = ["stepType", "typeOfLevel"]

    # TODO allow passing a LazyReferenceMapper as output?
    zarr_store_dict = {}
    zarr_store = dict_to_store(zarr_store_dict)
    zroot = zarr.open_group(store=zarr_store, zarr_format=2)

    aggregations: Dict[str, List] = defaultdict(list)
    aggregation_dims: Dict[str, Set] = defaultdict(set)

    unknown_counter = 0
    for msg_ind, group in enumerate(message_groups):
        assert group["version"] == 1

        gattrs = ujson.loads(group["refs"][".zattrs"])
        coordinates = gattrs["coordinates"].split(" ")

        # Find the data variable
        vname = None
        for key, entry in group["refs"].items():
            name = key.split("/")[0]
            if name not in [".zattrs", ".zgroup"] and name not in coordinates:
                vname = name
                break

        if vname is None:
            raise RuntimeError(
                f"Can not find a data var for msg# {msg_ind} in {group['refs'].keys()}"
            )

        if vname == "unknown":
            # To resolve unknown variables add custom grib tables.
            # https://confluence.ecmwf.int/display/UDOC/Creating+your+own+local+definitions+-+ecCodes+GRIB+FAQ
            # If you process the groups from a single file in order, you can use the msg# to compare with the
            # IDX file. The idx files message index is 1 based where the grib_tree message count is zero based
            logger.warning(
                "Dropping unknown variable in msg# %d. Compare with the grib idx file to help identify it"
                " and build an ecCodes local grib definitions file to fix it.",
                msg_ind,
            )
            unknown_counter += 1
            continue

        logger.debug("Processing vname: %s", vname)
        dattrs = ujson.loads(group["refs"][f"{vname}/.zattrs"])
        # filter order matters - it determines the hierarchy
        gfilters = {}
        for key in filters:
            attr_val = dattrs.get(f"GRIB_{key}")
            if attr_val is None:
                continue
            if attr_val == "unknown":
                logger.warning(
                    "Found 'unknown' attribute value for key %s in var %s of msg# %s",
                    key,
                    vname,
                    msg_ind,
                )
                # Use unknown as a group or drop it?

            gfilters[key] = attr_val

        zgroup = zroot.require_group(vname)
        if "name" not in zgroup.attrs:
            zgroup.attrs["name"] = dattrs.get("GRIB_name")

        for key, value in gfilters.items():
            if value:  # Ignore empty string and None
                # name the group after the attribute values: surface, instant, etc
                zgroup = zgroup.require_group(value)
                # Add an attribute to give context
                zgroup.attrs[key] = value

        # Set the coordinates attribute for the group
        zgroup.attrs["coordinates"] = " ".join(coordinates)
        # add to the list of groups to multi-zarr
        aggregations[zgroup.path].append(group)

        # keep track of the level coordinate variables and their values
        for key, entry in group["refs"].items():
            name = key.split("/")[0]
            if name == gfilters.get("typeOfLevel") and key.endswith("0"):
                if isinstance(entry, list):
                    entry = tuple(entry)
                aggregation_dims[zgroup.path].add(entry)

    concat_dims = ["time", "step"]
    identical_dims = ["longitude", "latitude"]
    for path in aggregations.keys():
        # Parallelize this step!
        catdims = concat_dims.copy()
        idims = identical_dims.copy()

        level_dimension_value_count = len(aggregation_dims.get(path, ()))
        level_group_name = path.split("/")[-1]
        if level_dimension_value_count == 0:
            logger.debug(
                "Path % has no value coordinate value associated with the level name %s",
                path,
                level_group_name,
            )
        elif level_dimension_value_count == 1:
            idims.append(level_group_name)
        elif level_dimension_value_count > 1:
            # The level name should be the last element in the path
            catdims.insert(3, level_group_name)

        logger.info(
            "%s calling MultiZarrToZarr with idims %s and catdims %s",
            path,
            idims,
            catdims,
        )

        mzz = MultiZarrToZarr(
            aggregations[path],
            remote_options=remote_options,
            concat_dims=catdims,
            identical_dims=idims,
        )
        group = mzz.translate()

        for key, value in group["refs"].items():
            if key not in [".zattrs", ".zgroup"]:
                zarr_store._store_dict[f"{path}/{key}"] = value

    # Force all stored values to decode as string, not bytes. String should be correct.
    # ujson will reject bytes values by default.
    # Using 'reject_bytes=False' one write would fail an equality check on read.
    zarr_dict = {
        key: (val.decode() if isinstance(val, bytes) else val)
        for key, val in zarr_store._store_dict.items()
    }
    # TODO handle other kerchunk reference spec versions?
    translate_refs_serializable(zarr_dict)
    result = dict(refs=zarr_dict, version=1)

    return result


def correct_hrrr_subhf_step(group: Dict) -> Dict:
    """
    Overrides the definition of the "step" variable.

    Sets the value equal to the `valid_time - time`
    in hours as a floating point value. This fixes issues with the HRRR SubHF grib2 step as read by
    cfgrib via scan_grib.
    The result is a deep copy, the original data is unmodified.

    Parameters
    ----------
    group: dict
        the zarr group store for a single grib message

    Returns
    -------
    dict: A new zarr store like dictionary for use as a reference filesystem mapper with zarr
    or xarray datatree
    """
    group = copy.deepcopy(group)
    group["refs"]["step/.zarray"] = (
        '{"chunks":[],"compressor":null,"dtype":"<f8","fill_value":"NaN","filters":null,"order":"C",'
        '"shape":[],"zarr_format":2}'
    )
    group["refs"]["step/.zattrs"] = (
        '{"_ARRAY_DIMENSIONS":[],"long_name":"time since forecast_reference_time",'
        '"standard_name":"forecast_period","units":"hours"}'
    )

    # add step to coords
    attrs = ujson.loads(group["refs"][".zattrs"])
    if "step" not in attrs["coordinates"]:
        attrs["coordinates"] += " step"
    group["refs"][".zattrs"] = ujson.dumps(attrs)

    fo = fsspec.filesystem("reference", fo=group, mode="r")
    fstore = fs_as_store(fo, read_only=True)
    xd = xarray.open_dataset(fstore, engine="zarr", consolidated=False)

    correct_step = xd.valid_time.values - xd.time.values

    assert correct_step.shape == ()
    step_float = correct_step.astype("timedelta64[s]").astype("float") / 3600.0
    step_bytes = step_float.tobytes()
    try:
        enocded_val = step_bytes.decode("ascii")
    except UnicodeDecodeError:
        enocded_val = (b"base64:" + base64.b64encode(step_bytes)).decode("ascii")

    group["refs"]["step/0"] = enocded_val

    return group


__all__ = [
    "scan_grib",
    "grib_tree",
    "correct_hrrr_subhf_step",
    "example_combine",
    "parse_grib_idx",
    "build_idx_grib_mapping",
    "map_from_index",
    "strip_datavar_chunks",
    "reinflate_grib_store",
    "extract_datatree_chunk_index",
    "AggregationType",
    "read_store",
    "write_store",
]



import numpy
import h5py
import hdf5plugin
import ujson
import numcodecs
import zarr
from fsspec.implementations.reference import ReferenceFileSystem


def _dict_add(a, b):
    c = a.copy()
    c.update(b)
    return c


compressors = dict(
    zstd=hdf5plugin.Zstd(),
    shuffle_zstd=_dict_add(
        dict(shuffle=True), hdf5plugin.Zstd()
    ),  # Test for two filters
    blosc=hdf5plugin.Blosc(),
    shuffle_blosc=_dict_add(dict(shuffle=True), hdf5plugin.Blosc()),
    lz4=hdf5plugin.LZ4(),
)

for c in compressors:
    f = h5py.File(f"hdf5_mini_{c}.h5", "w")
    f.create_dataset("data", (3, 2), dtype=numpy.int32, **compressors[c]).write_direct(
        numpy.arange(6, dtype=numpy.int32).reshape((3, 2))
    )
    f.close()

f = h5py.File(f"hdf5_mali_chunk.h5", "w")
f.create_dataset("data", (8,), dtype=numpy.int32, chunks=(7,), **hdf5plugin.Zstd())
f["data"][0:7] = numpy.arange(7, dtype=numpy.int32)
f["data"].id.write_direct_chunk((7,), numpy.array([7] + [0] * 6, dtype=numpy.int32), 1)
print(f["data"][:])
f.close()

f = h5py.File(f"hdf5_mali_chunk2.h5", "w")
f.create_dataset(
    "data", (8,), dtype=numpy.int32, chunks=(3,), shuffle=True, **hdf5plugin.Zstd()
)
f["data"].id.write_direct_chunk(
    (0,), numcodecs.Shuffle(4).encode(numpy.array([0, 1, 2], dtype=numpy.int32)), 2
)
f["data"].id.write_direct_chunk((3,), numpy.array([3, 4, 5], dtype=numpy.int32), 3)
f["data"][6:] = numpy.array(
    [
        6,
        7,
    ],
    dtype=numpy.int32,
)
print(f["data"][:])
f.close()

# from kerchunk.hdf import SingleHdf5ToZarr
# for c in compressors:
#     # if c=='lz4':
#     #     continue
#     with open(f'hdf5_mini_{c}.h5', 'rb') as f:
#         print(ujson.dumps(SingleHdf5ToZarr(f, None, inline_threshold=0, unsupported_inline_threshold=1).translate(), indent=4))
# with open(f'hdf5_mali_chunk.h5', 'rb') as f:
#     ref = SingleHdf5ToZarr(f, None).translate()
#     print(ujson.dumps(ref, indent=4))
# print(zarr.group(ReferenceFileSystem(ref, target=f'hdf5_mali_chunk.h5').get_mapper())['data'][:])
# with open(f'hdf5_mali_chunk2.h5', 'rb') as f:
#     ref = SingleHdf5ToZarr(f, None).translate()
#     print(ujson.dumps(ref, indent=4))
# print(zarr.group(ReferenceFileSystem(ref, target=f'hdf5_mali_chunk2.h5').get_mapper())['data'][:])
